#include <math.h>
#include "ladspa.h"
#include "wavedata.h"
#include "common.h"

/*****************************************************************************
 *
 * Description: Interpolates between p0 and n0 taking the previous (p1)
 *              and next (n1) points into account, using a 3rd order
 *              polynomial (aka cubic spline)
 *
 *   Arguments: interval    Normalised time interval between inteprolated
 *                           sample and p0
 *              p1, p0      Samples prior to interpolated one
 *              n0, n1      Samples following interpolated one
 *
 *     Returns: interpolated sample
 *
 *        Code: Adapted from Steve Harris' plugin code
 *              swh-plugins-0.2.7/ladspa-util.h::cube_interp
 *              http://plugin.org.uk/releases/0.2.7/
 *
 *****************************************************************************/
static inline LADSPA_Data interpolateCubic(LADSPA_Data interval,
                                           LADSPA_Data p1,
                                           LADSPA_Data p0,
                                           LADSPA_Data n0,
                                           LADSPA_Data n1)
{
    return p0 + 0.5f * interval * (n0 - p1 +
                       interval * (4.0f * n0 + 2.0f * p1 - 5.0f * p0 - n1 +
                       interval * (3.0f * (p0 - n0) - p1 + n1)));
}

/*****************************************************************************
 *
 * Description: Interpolates between p0 and n0 taking the previous (p1)
 *              and next (n1) points into account, using a 5th order
 *              polynomial.
 *
 *   Arguments: interval    Normalised time interval between inteprolated
 *                           sample and p0
 *              p1, p0      Samples prior to interpolated one
 *              n0, n1      Samples following interpolated one
 *
 *     Returns: interpolated sample
 *
 *        Code: http://www.musicdsp.org/archive.php?classid=5#62
 *
 *****************************************************************************/
static inline LADSPA_Data interpolateQuintic(LADSPA_Data interval,
                                             LADSPA_Data p1,
                                             LADSPA_Data p0,
                                             LADSPA_Data n0,
                                             LADSPA_Data n1)
{
    return p0 + 0.5f * interval * (n0 - p1 +
                       interval * (n0 - 2.0f * p0 + p1 +
                       interval * ( 9.0f * (n0 - p0) + 3.0f * (p1 - n1) +
                       interval * (15.0f * (p0 - n0) + 5.0f * (n1 - p1) +
                       interval * ( 6.0f * (n0 - p0) + 2.0f * (p1 - n1))))));
}

/*****************************************************************************
 *
 * Description: Get four samples from wavetable that surround given phase
 *
 *   Arguments: wtbl            Wavetable to extract sample from
 *              abs_freq        Absolute playback frequency
 *              phase           Phase (position) in wavetable
 *
 *     Returns: interval to use for interpolation
 *
 *       Notes: Cubic (or quintic) interpolation requires four consecutive
 *               samples for operation:
 *
 *                              phase
 *                                :
 *                    p1      p0  :   n0      n1
 *                    |       |   x   |       |
 *                            :   :
 *                            <-o->
 *                              :
 *                           interval
 *
 *              Phase values less than one make p0 the first sample in
 *               the table - p1 will be the last sample, as a previous
 *               sample does not exist. To avoid checking for this case,
 *               a copy of the last sample is stored before the first
 *               sample in each table.
 *              Likewise, if the phase means p0 is the last sample, n0
 *               and n1 will be the first and second samples respectively.
 *               Copies of these samples are stored after the last sample
 *               in each table.
 *
 *****************************************************************************/
static inline LADSPA_Data getSamples(Wavetable *wtbl,
                                     LADSPA_Data abs_freq,
                                     LADSPA_Data phase,
                                     LADSPA_Data *p1,
                                     LADSPA_Data *p0,
                                     LADSPA_Data *n0,
                                     LADSPA_Data *n1)
{
    unsigned long count;
    LADSPA_Data *samples_base;
    LADSPA_Data *samples_extra;
    LADSPA_Data phase_f;
    LADSPA_Data xfade_scale;
    LADSPA_Data p1_b, p0_b, n0_b, n1_b;
    LADSPA_Data p1_e, p0_e, n0_e, n1_e;
    long index;

    count = wtbl->count;
    samples_base = wtbl->samples_base;
    samples_extra = wtbl->samples_extra;

/* Scale phase to map to position in wavetable */
    phase *= wtbl->phase_scale_factor;

/* Phase is between p0 and n0 */
    phase_f = (LADSPA_Data)floorf((float)phase);

/* Get position of second contributing sample (p0) */
    index = (long)phase_f;
    index %= count;

    switch (wtbl->fade) {
    case YES:
    /* Previous two samples */
        p1_b = samples_base[index];
        p1_e = samples_extra[index];
        index++;
        p0_b = samples_base[index];
        p0_e = samples_extra[index];
        index++;
    /* Next two samples */
        n0_b = samples_base[index];
        n0_e = samples_extra[index];
        index++;
        n1_b = samples_base[index];
        n1_e = samples_extra[index];

        xfade_scale = (wtbl->max_frequency - abs_freq) * wtbl->range_scale_factor;
        xfade_scale = xfade_scale < 0.0f ? 0.0f : xfade_scale;
        xfade_scale = xfade_scale > 1.0f ? 1.0f : xfade_scale;
        *p1 = p1_b + xfade_scale * p1_e;
        *p0 = p0_b + xfade_scale * p0_e;
        *n0 = n0_b + xfade_scale * n0_e;
        *n1 = n1_b + xfade_scale * n1_e;
        break;
    case NO:
    /* Previous two samples */
        *p1 = samples_base[index++];
        *p0 = samples_base[index++];
    /* Next two samples */
        *n0 = samples_base[index++];
        *n1 = samples_base[index];
        break;
    case OUT:
        xfade_scale = (wtbl->max_frequency - abs_freq) * wtbl->range_scale_factor;
        xfade_scale = xfade_scale < 0.0f ? 0.0f : xfade_scale;
        xfade_scale = xfade_scale > 1.0f ? 1.0f : xfade_scale;
    /* Previous two samples */
        *p1 = xfade_scale * samples_base[index++];
        *p0 = xfade_scale * samples_base[index++];
    /* Next two samples */
        *n0 = xfade_scale * samples_base[index++];
        *n1 = xfade_scale * samples_base[index];
        break;
    }

/* Return interval for interpolation */
    return phase - phase_f;
}

/*****************************************************************************
 *
 * Description: Get wavetable to use for playback frequency.
 *
 *   Arguments: wdat            Wavedata object (contains all table info)
 *              abs_freq        Absolute playback frequency
 *
 *     Returns: Wavetable to use
 *              NULL if frequency exceeds Nyquist
 *
 *       Notes: The lookup vector used to determine the wavetable
 *               is indexed by harmonic number.
 *
 *              The maximum playback frequency for a wavetable is
 *               determined by its harmonic content and the sample rate,
 *               and equals sample_rate / 2 / max_harmonic_in_table.
 *
 *****************************************************************************/
static inline Wavetable* getTable(Wavedata *wdat,
                                  LADSPA_Data abs_freq)
{
    long harmonic;

/* Frequencies near zero always use low freq wavetable */
/* This catches any divide by zero's */
    if (abs_freq < 0.125f) {
        return wdat->wtbl_lf;
    }

/* Get highest harmonic possible in frequency */
    harmonic = (long)(wdat->sample_rate / 2.0f / abs_freq);

/* If it exceeds the maximum lookup bound use low freq table */
    if (harmonic > wdat->lookup_max) {
        return wdat->wtbl_lf;
    }

/* Harmonic == 0 means frequency exceeds Nyquist */
    if (harmonic == 0) {
    /* No table to play */
        return NULL;
    }

    harmonic -= wdat->lookup_min;
    if (harmonic < 0) {
        if (abs_freq > wdat->wtbl_fadeout->min_frequency) {
            return wdat->wtbl_fadeout;
        }
        return wdat->wtbl_hf;
    }

/* Is within bounds */
    return wdat->wavetables[wdat->lookup[harmonic]];
}
