/*
    wdatutil.h - Code to generate c header files containing pre-calculated
                 wavetables.

    Copyright (C) 2002  Mike Rawes

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef MYK_WDAT_TOOLS_H
#define MYK_WDAT_TOOLS_H

#include <stdio.h>
#include "ladspa.h"

#define WAVE_TYPE_COUNT         3

extern char *wave_names[];
extern char *wave_descriptions[];
extern unsigned long wave_first_harmonics[];
extern unsigned long wave_harmonic_intervals[];

/* Get actual maximum harmonic from given harmonic, h, and wavetype, w */
#define ACTUAL_HARM(h,w) h - (h - wave_first_harmonics[w]) % wave_harmonic_intervals[w]
/* Get minimum harmonic content in given wavetype, w */
#define MIN_HARM(w) wave_first_harmonics[w]
/* Get minimum extra harmonic content possible in given wavetype, w */
#define MIN_EXTRA_HARM(w) wave_harmonic_intervals[w]
/* Get frequency from MIDI note, n */
#define FREQ_FROM_NOTE(n) 6.875f * powf(2.0f, (float)(n + 3) / 12.0f)
/* Get max harmonic from given frequency, f, at sample rate, r */
#define HARM_FROM_FREQ(f,r) (unsigned long)((float)sample_rate / freq / 2.0f)

/*
 * This is for creating c headers containing pre-calculated wavetable data, and
 * is driven by the Wavegen program.
 *
 * The header is then included by wavedata.c, which contains functions to read
 * the data into the Wavedata structure used by the plugins.
 */

/*
 * A single wavetable will have a range of pitches at which their samples
 * may be played back.
 *
 * The upper bound is determined by the maximum harmonic present in the
 * waveform - above this frequency, the higher harmonics will alias.
 *
 * The lower bound is chosen to be the higher bound of the previous wavetable
 * (or a pre-defined limit if there is no such table).
 */
typedef struct
{
    unsigned long sample_count;             /* Number of samples in table */
    unsigned long max_harmonic_base;        /* Highest harmonic in 'base' samples */
    unsigned long max_harmonic_extra;       /* Highest harmonic in 'extra' samples */
    LADSPA_Data *samples_base;              /* Base samples - these are for playback across
                                                the whole frequency range */
    LADSPA_Data *samples_extra;             /* Extra samples - these are played back at
                                                increasing amplitude as pitch is reduced */
} Wtblfile;

/* Wavedata file structure */
typedef struct
{
    unsigned long sample_rate;              /* Sample rate of generated data */
    unsigned long number_of_wavetables;     /* Number of wavetables in wavedata */
    unsigned long lowest_harmonic;          /* Lowest harmonic */
    unsigned long highest_harmonic;         /* Highest harmonic */
    Wtblfile **wavetables;                  /* Wavetables */
} Wdatfile;

#ifdef __cplusplus
extern "C" {
#endif

/*******************************************************************************
 *            Name: createWavedataFile
 *
 *     Description: Initialise (malloc) wavedata file for holding wavedata
 *                  before writing
 *
 *       Arguments: sample_rate       Sample rate to use when generating data
 *                  lowest_harmonic   Lowest harmonic
 *                  highest_harmonic  Highest harmonic
 *
 *         Returns: Pointer to new wavedata structure on success
 *                  NULL otherwise
 *
 *           Notes: No wavetables are allocated. Use addWavetable.
 ******************************************************************************/
Wdatfile *createWavedataFile(unsigned long sample_rate,
                             unsigned long lowest_harmonic,
                             unsigned long highest_harmonic);

/*******************************************************************************
 *            Name: destroyWavedataFile
 *
 *     Description: Destroy wavedata file object, and free allocated memory
 *
 *       Arguments: wdatfile        Wavedata file object
 *
 *           Notes: wdatfile should have been initialised using createWavedataFile
 *                  Null pointer checking is not done before free()
 ******************************************************************************/
void destroyWavedataFile(Wdatfile *wdatfile);

/*******************************************************************************
 *            Name: addWavetable
 *
 *     Description: Add new wavetable information to wavedata file object
 *
 *       Arguments: wdatfile            Wavedata file object to add to
 *                  sample_count        Number of samples in wavetable
 *                  max_harmonic_base   Maximum harmonic in base data
 *                  max_harmonic_extra  Maximum harmonic in extra data
 *
 *         Returns: TRUE on success
 *                  FALSE otherwise
 ******************************************************************************/
int addWavetable(Wdatfile *wdatfile,
                 unsigned long sample_count,
                 unsigned long max_harmonic_base,
                 unsigned long max_harmonic_extra);

/*******************************************************************************
 *            Name: generateWave
 *
 *     Description: Initialise wavetable data with a waveform generated from
 *                  Fourier series with a maximum harmonic.
 *
 *       Arguments: wtblfile        The wavetable to generate data for
 *                  wavetype_index  Index of wavetype to generate
 *                  gibbs_comp      Compensation for Gibbs' effect:
 *                                    0.0: none (wave will overshoot)
 *                                    1.0: full (wave will not overshoot)
 *
 *           Notes: Compensation for Gibbs' Effect will reduce the degree
 *                   of overshoot at the transitions. A value of 1.0 will
 *                   pretty much eliminate it, and cause the wave to be
 *                   normalised.
 ******************************************************************************/
void generateWave(Wtblfile *wtblfile,
                  int wavetype_index,
                  float gibbs_comp);

/*******************************************************************************
 *            Name: writeWavedataFile
 *
 *     Description: Write wavedata as a c header file
 *
 *       Arguments: wdatfile    Wavedata file to write
 *                  wdat_fp     Pointer to output file
 *                  prefix      Prefix to use in declarations. If this is null
 *                               declarations are prefixed with 'wdat'.
 *
 *         Returns: TRUE on success
 *                  FALSE otherwise
 ******************************************************************************/
int writeWavedataFile(Wdatfile *wdatfile,
                      FILE *wdat_fp,
                      char *prefix);

#ifdef __cplusplus
} /* extern "C" { */
#endif

#endif /* MYK_WDAT_TOOLS_H */
