/*
    lp4pole.so.c - A LADSPA plugin simulating a 4 pole low pass resonant
                   filter.

    Originally appeared in the arts softsynth by Stefan Westerfeld:
    http://www.arts-project.org/

    First ported to LADSPA by Reiner Klenk (pdq808[at]t-online.de)

    Tuning and stability issues (output NaN) and additional audio-rate
    variant added by Mike Rawes (mike_rawes[at]yahoo.co.uk)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdlib.h>
#include <string.h>
#include <ladspa.h>

/*
 * Only two variants - there's not much ground gained by having
 * Cutoff/Resonance with Audio/Control ports
 */

#define LP4POLE_BASE_ID                1671
#define LP4POLE_VARIANT_COUNT          2

#define LP4POLE_CUTOFF                 0
#define LP4POLE_RESONANCE              1
#define LP4POLE_INPUT                  2
#define LP4POLE_OUTPUT                 3

LADSPA_Descriptor **lp4pole_descriptors = NULL;

typedef struct {
    LADSPA_Data *cutoff;
    LADSPA_Data *resonance;
    LADSPA_Data *input;
    LADSPA_Data *output;
    LADSPA_Data in1;
    LADSPA_Data in2;
    LADSPA_Data in3;
    LADSPA_Data in4;
    LADSPA_Data inv_nyquist;
    LADSPA_Data nyquist;
    LADSPA_Data out1;
    LADSPA_Data out2;
    LADSPA_Data out3;
    LADSPA_Data out4;
} Lp4pole;

const LADSPA_Descriptor *ladspa_descriptor(unsigned long index)
{
    if (index < LP4POLE_VARIANT_COUNT) {
        return lp4pole_descriptors[index];
    } else {
        return NULL;
    }
}

void cleanupLp4pole(LADSPA_Handle instance)
{
    free(instance);
}

void connectPortLp4pole(LADSPA_Handle instance,
                        unsigned long port,
                        LADSPA_Data *data)
{
    Lp4pole *plugin;

    plugin = (Lp4pole *)instance;
    switch (port) {
    case LP4POLE_CUTOFF:
        plugin->cutoff = data;
        break;
    case LP4POLE_RESONANCE:
        plugin->resonance = data;
        break;
    case LP4POLE_INPUT:
        plugin->input = data;
        break;
    case LP4POLE_OUTPUT:
        plugin->output = data;
        break;
    }
}

LADSPA_Handle instantiateLp4pole(const LADSPA_Descriptor *descriptor,
                                 unsigned long sample_rate)
{
    Lp4pole *plugin = (Lp4pole *)malloc(sizeof(Lp4pole));
    LADSPA_Data in1;
    LADSPA_Data in2;
    LADSPA_Data in3;
    LADSPA_Data in4;
    LADSPA_Data inv_nyquist;
    LADSPA_Data nyquist;
    LADSPA_Data out1;
    LADSPA_Data out2;
    LADSPA_Data out3;
    LADSPA_Data out4;

    nyquist = (LADSPA_Data)sample_rate / 2.0f;
    inv_nyquist = 1.0f / nyquist;

    in1 = 0.0f;
    in2 = 0.0f;
    in3 = 0.0f;
    in4 = 0.0f;

    out1 = 0.0f;
    out2 = 0.0f;
    out3 = 0.0f;
    out4 = 0.0f;

    plugin->in1 = in1;
    plugin->in2 = in2;
    plugin->in3 = in3;
    plugin->in4 = in4;
    plugin->inv_nyquist = inv_nyquist;
    plugin->nyquist = nyquist;
    plugin->out1 = out1;
    plugin->out2 = out2;
    plugin->out3 = out3;
    plugin->out4 = out4;

    return (LADSPA_Handle)plugin;
}

#undef buffer_write
#undef RUN_ADDING
#undef RUN_REPLACING

#define buffer_write(b, v) (b = v)
#define RUN_ADDING    0
#define RUN_REPLACING 1

void runLp4pole_faraia_oa(LADSPA_Handle instance,
                          unsigned long sample_count)
{
    Lp4pole *plugin = (Lp4pole *)instance;

    /* Cutoff Frequency (array of floats of length sample_count) */
    LADSPA_Data *cutoff = plugin->cutoff;

    /* Resonance (array of floats of length sample_count) */
    LADSPA_Data *resonance = plugin->resonance;

    /* Input (array of floats of length sample_count) */
    LADSPA_Data *input = plugin->input;

    /* Output (pointer to float value) */
    LADSPA_Data *output = plugin->output;

    LADSPA_Data in1 = plugin->in1;
    LADSPA_Data in2 = plugin->in2;
    LADSPA_Data in3 = plugin->in3;
    LADSPA_Data in4 = plugin->in4;
    LADSPA_Data inv_nyquist = plugin->inv_nyquist;
    LADSPA_Data nyquist = plugin->nyquist;
    LADSPA_Data out1 = plugin->out1;
    LADSPA_Data out2 = plugin->out2;
    LADSPA_Data out3 = plugin->out3;
    LADSPA_Data out4 = plugin->out4;

    LADSPA_Data in;
    LADSPA_Data co;
    LADSPA_Data res;
    LADSPA_Data tuning;
    LADSPA_Data f;
    LADSPA_Data fsqd;
    LADSPA_Data fb;
    unsigned long s;

    for (s = 0; s < sample_count; s++) {
        co = cutoff[s];
        res = resonance[s];
        in = input[s];

    /* Find tuning - Magic numbers found empirically :) */
        tuning = co < 8844.0f ? 3.13 - (co * 1.7695983e-04f) : 1.56503274f;
        f = co * tuning * inv_nyquist;

    /* Clamp result */
        f = f < inv_nyquist ? inv_nyquist : f;
        f = f > 1.16f ? 1.16f : f;

        fsqd = f * f;
        f = 1.0f - f;

        fb = res * (1.0f - 0.15f * fsqd);

        in -= out4 * fb;
        in *= 0.35013f * fsqd * fsqd;

        out1 = in + 0.3f * in1 + f * out1; /* Pole 1 */
        in1  = in;
        out2 = out1 + 0.3f * in2 + f * out2;  /* Pole 2 */
        in2  = out1;
        out3 = out2 + 0.3f * in3 + f * out3;  /* Pole 3 */
        in3  = out2;
        out4 = out3 + 0.3f * in4 + f * out4;  /* Pole 4 */
        in4  = out3;

        buffer_write(output[s], out4);
    }

    plugin->in1 = in1;
    plugin->in2 = in2;
    plugin->in3 = in3;
    plugin->in4 = in4;
    plugin->out1 = out1;
    plugin->out2 = out2;
    plugin->out3 = out3;
    plugin->out4 = out4;
}

void runLp4pole_fcrcia_oa(LADSPA_Handle instance,
                          unsigned long sample_count)
{
    Lp4pole *plugin = (Lp4pole *)instance;

    /* Cutoff Frequency (float value) */
    LADSPA_Data cutoff = *(plugin->cutoff);

    /* Resonance (float value) */
    LADSPA_Data resonance = *(plugin->resonance);

    /* Input (array of floats of length sample_count) */
    LADSPA_Data *input = plugin->input;

    /* Output (pointer to float value) */
    LADSPA_Data *output = plugin->output;
    LADSPA_Data in1 = plugin->in1;
    LADSPA_Data in2 = plugin->in2;
    LADSPA_Data in3 = plugin->in3;
    LADSPA_Data in4 = plugin->in4;
    LADSPA_Data inv_nyquist = plugin->inv_nyquist;
    LADSPA_Data nyquist = plugin->nyquist;
    LADSPA_Data out1 = plugin->out1;
    LADSPA_Data out2 = plugin->out2;
    LADSPA_Data out3 = plugin->out3;
    LADSPA_Data out4 = plugin->out4;

    LADSPA_Data in;
    LADSPA_Data tuning;
    LADSPA_Data f;
    LADSPA_Data fsqd;
    LADSPA_Data fpow4;
    LADSPA_Data fb;
    unsigned long s;

/* Find tuning - Magic numbers found empirically :) */
    tuning = cutoff < 8844.0f ? 3.13f - (cutoff * 1.7695983e-04f) : 1.56503274f;
    f = cutoff * tuning * inv_nyquist;

/* Clamp result */
    f = f < inv_nyquist ? inv_nyquist : f;
    f = f > 1.16f ? 1.16f : f;

    fsqd = f*f;
    fpow4 = fsqd*fsqd;

    fb = resonance * (1.0f - 0.15f * fsqd);

    f = 1.0f - f;

    for (s = 0; s < sample_count; s++) {
        in = input[s];
        in -= out4 * fb;
        in *= 0.35013f * fpow4;

        out1 = in + 0.3f * in1 + f * out1; /* Pole 1 */
        in1  = in;
        out2 = out1 + 0.3f * in2 + f * out2;  /* Pole 2 */
        in2  = out1;
        out3 = out2 + 0.3f * in3 + f * out3;  /* Pole 3 */
        in3  = out2;
        out4 = out3 + 0.3f * in4 + f * out4;  /* Pole 4 */
        in4  = out3;

        buffer_write(output[s], out4);
    }

    plugin->in1 = in1;
    plugin->in2 = in2;
    plugin->in3 = in3;
    plugin->in4 = in4;
    plugin->out1 = out1;
    plugin->out2 = out2;
    plugin->out3 = out3;
    plugin->out4 = out4;
}

void _init()
{
    char **port_names;
    LADSPA_PortDescriptor *port_descriptors;
    LADSPA_PortRangeHint *port_range_hints;
    LADSPA_Descriptor *descriptor;
    int i;

    const char *labels[] = {"lp4pole_faraia_oa",
                            "lp4pole_fcrcia_oa"};
    LADSPA_PortDescriptor cutoff_port_descriptors[] = {LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO,
                                                       LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL};
    LADSPA_PortDescriptor resonance_port_descriptors[] = {LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO,
                                                          LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL};
    LADSPA_PortDescriptor input_port_descriptors[] = {LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO,
                                                      LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO};
    LADSPA_PortDescriptor output_port_descriptors[] = {LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
                                                       LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO};

    void (*run_functions[])(LADSPA_Handle,
                            unsigned long) = {runLp4pole_faraia_oa,
                                              runLp4pole_fcrcia_oa};
    lp4pole_descriptors = (LADSPA_Descriptor **)calloc(LP4POLE_VARIANT_COUNT, sizeof(LADSPA_Descriptor));

    if (lp4pole_descriptors) {
        for (i = 0; i < LP4POLE_VARIANT_COUNT; i++) {
            lp4pole_descriptors[i] = (LADSPA_Descriptor *)malloc(sizeof(LADSPA_Descriptor));
            descriptor = lp4pole_descriptors[i];
            if (descriptor) {
                descriptor->UniqueID = LP4POLE_BASE_ID + i;
                descriptor->Label = strdup(labels[i]);
                descriptor->Properties = LADSPA_PROPERTY_HARD_RT_CAPABLE;
                descriptor->Name = strdup("4 Pole Low-Pass Filter with Resonance");
                descriptor->Maker = strdup("Mike Rawes <mike_rawes[at]yahoo.co.uk>");
                descriptor->Copyright = strdup("GPL");

                descriptor->PortCount = 4;

                port_descriptors = (LADSPA_PortDescriptor *)calloc(4, sizeof(LADSPA_PortDescriptor));
                descriptor->PortDescriptors = (const LADSPA_PortDescriptor *)port_descriptors;

                port_range_hints = (LADSPA_PortRangeHint *)calloc(4, sizeof(LADSPA_PortRangeHint));
                descriptor->PortRangeHints = (const LADSPA_PortRangeHint *)port_range_hints;

                port_names = (char **)calloc(4, sizeof(char*));
                descriptor->PortNames = (const char **)port_names;

                /* Parameters for Cutoff Frequency */
                port_descriptors[LP4POLE_CUTOFF] = cutoff_port_descriptors[i];
                port_names[LP4POLE_CUTOFF] = strdup("Cutoff Frequency");
                port_range_hints[LP4POLE_CUTOFF].HintDescriptor = LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_SAMPLE_RATE | LADSPA_HINT_LOGARITHMIC;
                port_range_hints[LP4POLE_CUTOFF].LowerBound = 0.0f;
                port_range_hints[LP4POLE_CUTOFF].UpperBound = 0.5f;

                /* Parameters for Resonance */
                port_descriptors[LP4POLE_RESONANCE] = resonance_port_descriptors[i];
                port_names[LP4POLE_RESONANCE] = strdup("Resonance");
                port_range_hints[LP4POLE_RESONANCE].HintDescriptor = LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
                port_range_hints[LP4POLE_RESONANCE].LowerBound = 0.0f;
                port_range_hints[LP4POLE_RESONANCE].UpperBound = 4.0f;

                /* Parameters for Input */
                port_descriptors[LP4POLE_INPUT] = input_port_descriptors[i];
                port_names[LP4POLE_INPUT] = strdup("Input");
                port_range_hints[LP4POLE_INPUT].HintDescriptor = 0;

                /* Parameters for Output */
                port_descriptors[LP4POLE_OUTPUT] = output_port_descriptors[i];
                port_names[LP4POLE_OUTPUT] = strdup("Output");
                port_range_hints[LP4POLE_OUTPUT].HintDescriptor = 0;

                descriptor->activate = NULL;
                descriptor->cleanup = cleanupLp4pole;
                descriptor->connect_port = connectPortLp4pole;
                descriptor->deactivate = NULL;
                descriptor->instantiate = instantiateLp4pole;
                descriptor->run = run_functions[i];
                descriptor->run_adding = NULL;
                descriptor->set_run_adding_gain = NULL;
            }
        }
    }
}

void _fini() 
{
    LADSPA_Descriptor *descriptor;
    int i, j;

    if (lp4pole_descriptors) {
        for (i = 0; i < LP4POLE_VARIANT_COUNT; i++) {
            descriptor = lp4pole_descriptors[i];
            if (descriptor) {
                free((char *)descriptor->Label);
                free((char *)descriptor->Name);
                free((char *)descriptor->Maker);
                free((char *)descriptor->Copyright);
                free((LADSPA_PortDescriptor *)descriptor->PortDescriptors);

                for (j = 0; j < descriptor->PortCount; j++) {
                    free((char *)(descriptor->PortNames[j]));
                }
                free((char **)descriptor->PortNames);
                free((LADSPA_PortRangeHint *)descriptor->PortRangeHints);
                free(descriptor);
            }
        }
        free(lp4pole_descriptors);
    }
}
