/*
    pulse.so.c - A LADSPA plugin to generate a bandlimited variable pulse
                 waveform

    Copyright (C) 2002  Mike Rawes

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "ladspa.h"
#include "wavedata.h"
#include "interpolate.h"

/* This contains all the generated table data */
#include "sawtooth_data.h"

#define INTERPOLATE(i,a,b,c,d) interpolateCubic(i,a,b,c,d)

#define PULSE_BASE_ID                  1645
#define PULSE_VARIANT_COUNT            4

#define PULSE_FREQUENCY                0
#define PULSE_PULSEWIDTH               1
#define PULSE_OUTPUT                   2

LADSPA_Descriptor **pulse_descriptors = NULL;

typedef struct {
    LADSPA_Data *frequency;
    LADSPA_Data *pulsewidth;
    LADSPA_Data *output;
    Wavedata *wdat;
    LADSPA_Data srate;
    LADSPA_Data phase;
} Pulse;

/*****************************************************************************
 *
 * LADSPA Plugin code
 *
 *****************************************************************************/

const LADSPA_Descriptor *ladspa_descriptor(unsigned long index)
{
    if (index < 4) {
        return pulse_descriptors[index];
    } else {
        return NULL;
    }
}

void cleanupPulse(LADSPA_Handle instance)
{
    Pulse *plugin = (Pulse *)instance;

    destroyWavedata(plugin->wdat);
    free(instance);
}

void connectPortPulse(LADSPA_Handle instance,
                      unsigned long port,
                      LADSPA_Data *data)
{
    Pulse *plugin = (Pulse *)instance;

    switch (port) {
    case PULSE_FREQUENCY:
        plugin->frequency = data;
        break;
    case PULSE_PULSEWIDTH:
        plugin->pulsewidth = data;
        break;
    case PULSE_OUTPUT:
        plugin->output = data;
        break;
    }
}

LADSPA_Handle instantiatePulse(const LADSPA_Descriptor *descriptor,
                               unsigned long sample_rate)
{
    Pulse *plugin = (Pulse *)malloc(sizeof(Pulse));

    Wdatheader saw_data;

/* Pass data from header to wavedata initialisation */
    saw_data.number_of_wavetables = wdat_number_of_wavetables;
    saw_data.lowest_harmonic = wdat_lowest_harmonic;
    saw_data.highest_harmonic = wdat_highest_harmonic;
    saw_data.max_harmonics = wdat_max_harmonics;
    saw_data.sample_counts = wdat_sample_counts;
    saw_data.sample_data_base = wdat_sample_data_base;
    saw_data.sample_data_extra = wdat_sample_data_extra;

    plugin->wdat = initWavedata(&saw_data, sample_rate);
    if (!plugin->wdat) {
        return NULL;
    }
    plugin->srate = plugin->wdat->sample_rate;
    plugin->phase = 0.0f;

    return (LADSPA_Handle)plugin;
}

#undef buffer_write
#undef RUN_ADDING
#undef RUN_REPLACING

#define buffer_write(b, v) (b = v)
#define RUN_ADDING    0
#define RUN_REPLACING 1

void runPulse_fapa_oa(LADSPA_Handle instance,
                      unsigned long sample_count)
{
    Pulse *plugin = (Pulse *)instance;

    /* Frequency (array of LADSPA_Data of length sample_count) */
    LADSPA_Data *frequency = plugin->frequency;

    /* Pulse Width (array of LADSPA_Data of length sample_count) */
    LADSPA_Data *pulsewidth = plugin->pulsewidth;

    /* Output (pointer to LADSPA_Data value) */
    LADSPA_Data *output = plugin->output;

    /* Instance data */
    Wavedata *wdat = plugin->wdat;
    LADSPA_Data phase = plugin->phase;
    LADSPA_Data srate = plugin->srate;

    Wavetable *wtbl;
    LADSPA_Data freq;
    LADSPA_Data abs_freq;
    LADSPA_Data pwidth;
    LADSPA_Data shifted_phase;
    LADSPA_Data interval;
    LADSPA_Data p1, p0, n0, n1;
    LADSPA_Data sample;
    unsigned long s;

    for (s = 0; s < sample_count; s++) {
        freq = frequency[s];
        abs_freq = (LADSPA_Data)fabsf((float)freq);

        pwidth = pulsewidth[s];

    /* Lookup which table to use from frequency */
        wtbl = getTable(wdat, abs_freq);
        if (wtbl) {
        /* Get samples from sawtooth and phase shifted inverted sawtooth */
            interval = getSamples(wtbl, abs_freq, phase, &p1, &p0, &n0, &n1);
            sample = INTERPOLATE(interval, p1, p0, n0, n1);

            pwidth = pwidth < 0.0f ? 0.0f : pwidth;
            pwidth = pwidth > 1.0f ? 1.0f : pwidth;
            shifted_phase = phase + pwidth * srate;

            interval = getSamples(wtbl, abs_freq, shifted_phase, &p1, &p0, &n0, &n1);
            sample -= INTERPOLATE(interval, p1, p0, n0, n1);

        /* Offset to center DC = 0 */
            buffer_write(output[s], sample + 1.0f - (2.0f * pwidth));

            phase += freq;
            if (phase < srate) {
                phase += srate;
            } else if (phase > srate) {
                phase -= srate;
            }
        } else {
        /* No table found (implies frequency exceeds nyquist) */
            buffer_write(output[s], 0.0f);
        }
    }

    plugin->phase = phase;
}

void runPulse_fapc_oa(LADSPA_Handle instance,
                      unsigned long sample_count)
{
    Pulse *plugin = (Pulse *)instance;

    /* Frequency (array of LADSPA_Data of length sample_count) */
    LADSPA_Data *frequency = plugin->frequency;

    /* Pulse Width (LADSPA_Data value) */
    LADSPA_Data pulsewidth = *(plugin->pulsewidth);

    /* Output (pointer to LADSPA_Data value) */
    LADSPA_Data *output = plugin->output;

    /* Instance data */
    Wavedata *wdat = plugin->wdat;
    LADSPA_Data phase = plugin->phase;
    LADSPA_Data srate = plugin->srate;

    Wavetable *wtbl;
    LADSPA_Data freq;
    LADSPA_Data abs_freq;
    LADSPA_Data shifted_phase;
    LADSPA_Data dc_shift = 1.0 - (2.0 * pulsewidth);
    LADSPA_Data interval;
    LADSPA_Data p1, p0, n0, n1;
    LADSPA_Data sample;
    unsigned long s;

    for (s = 0; s < sample_count; s++) {
        freq = frequency[s];
        abs_freq = (LADSPA_Data)fabsf((float)freq);

    /* Lookup which table to use from frequency */
        wtbl = getTable(wdat, abs_freq);
        if (wtbl) {
        /* Get samples from sawtooth and phase shifted inverted sawtooth */
            interval = getSamples(wtbl, abs_freq, phase, &p1, &p0, &n0, &n1);
            sample = INTERPOLATE(interval, p1, p0, n0, n1);

            pulsewidth = pulsewidth < 0.0f ? 0.0f : pulsewidth;
            pulsewidth = pulsewidth > 1.0f ? 1.0f : pulsewidth;
            shifted_phase = phase + pulsewidth * srate;

            interval = getSamples(wtbl, abs_freq, shifted_phase, &p1, &p0, &n0, &n1);
            sample -= INTERPOLATE(interval, p1, p0, n0, n1);

        /* Offset to center on DC = 0 */
            buffer_write(output[s], sample + dc_shift);

            phase += freq;
            if (phase < srate) {
                phase += srate;
            } else if (phase > srate) {
                phase -= srate;
            }
        } else {
        /* No table found (implies frequency exceeds nyquist) */
            buffer_write(output[s], 0.0f);
        }
    }

    plugin->phase = phase;
}

void runPulse_fcpa_oa(LADSPA_Handle instance,
                      unsigned long sample_count)
{
    Pulse *plugin = (Pulse *)instance;

    /* Frequency (LADSPA_Data value) */
    LADSPA_Data frequency = *(plugin->frequency);

    /* Pulse Width (array of LADSPA_Data of length sample_count) */
    LADSPA_Data *pulsewidth = plugin->pulsewidth;

    /* Output (pointer to LADSPA_Data value) */
    LADSPA_Data *output = plugin->output;

    /* Instance data */
    Wavedata *wdat = plugin->wdat;
    LADSPA_Data phase = plugin->phase;
    LADSPA_Data srate = plugin->srate;

    Wavetable *wtbl;
    LADSPA_Data abs_freq = (LADSPA_Data)fabsf((float)frequency);
    LADSPA_Data pwidth;
    LADSPA_Data shifted_phase;
    LADSPA_Data interval;
    LADSPA_Data p1, p0, n0, n1;
    LADSPA_Data sample;
    unsigned long s;

    wtbl = getTable(wdat, abs_freq);
    if (wtbl) {
        for (s = 0; s < sample_count; s++) {
            pwidth = pulsewidth[s];

        /* Get samples from sawtooth and phase shifted inverted sawtooth */
            interval = getSamples(wtbl, abs_freq, phase, &p1, &p0, &n0, &n1);
            sample = INTERPOLATE(interval, p1, p0, n0, n1);

            pwidth = pwidth < 0.0f ? 0.0f : pwidth;
            pwidth = pwidth > 1.0f ? 1.0f : pwidth;
            shifted_phase = phase + pwidth * srate;

            interval = getSamples(wtbl, abs_freq, shifted_phase, &p1, &p0, &n0, &n1);
            sample -= INTERPOLATE(interval, p1, p0, n0, n1);

        /* Offset to center DC = 0 */
            buffer_write(output[s], sample + 1.0f - (2.0f * pwidth));

            phase += frequency;
            if (phase < srate) {
                phase += srate;
            } else if (phase > srate) {
                phase -= srate;
            }
        }
    } else {
    /* No table found (implies frequency exceeds nyquist) */
        for (s = 0; s < sample_count; s++) {
            buffer_write(output[s], 0.0f);
        }
    }

    plugin->phase = phase;
}

void runPulse_fcpc_oa(LADSPA_Handle instance,
                      unsigned long sample_count)
{
    Pulse *plugin = (Pulse *)instance;

    /* Frequency (LADSPA_Data value) */
    LADSPA_Data frequency = *(plugin->frequency);

    /* Pulse Width (LADSPA_Data value) */
    LADSPA_Data pulsewidth = *(plugin->pulsewidth);

    /* Output (pointer to LADSPA_Data value) */
    LADSPA_Data *output = plugin->output;

    /* Instance data */
    Wavedata *wdat = plugin->wdat;
    LADSPA_Data phase = plugin->phase;
    LADSPA_Data srate = plugin->srate;

    Wavetable *wtbl;
    LADSPA_Data abs_freq = (LADSPA_Data)fabsf((float)frequency);
    LADSPA_Data shifted_phase;
    LADSPA_Data dc_shift = 1.0f - (2.0f * pulsewidth);
    LADSPA_Data interval;
    LADSPA_Data p1, p0, n0, n1;
    LADSPA_Data sample;
    unsigned long s;

    wtbl = getTable(wdat, abs_freq);
    if (wtbl) {
        for (s = 0; s < sample_count; s++) {
        /* Get samples from sawtooth and phase shifted inverted sawtooth */
            interval = getSamples(wtbl, abs_freq, phase, &p1, &p0, &n0, &n1);
            sample = INTERPOLATE(interval, p1, p0, n0, n1);

            pulsewidth = pulsewidth < 0.0f ? 0.0f : pulsewidth;
            pulsewidth = pulsewidth > 1.0f ? 1.0f : pulsewidth;
            shifted_phase = phase + pulsewidth * srate;

            interval = getSamples(wtbl, abs_freq, shifted_phase, &p1, &p0, &n0, &n1);
            sample -= INTERPOLATE(interval, p1, p0, n0, n1);

        /* Offset to center DC = 0 */
            buffer_write(output[s], sample + dc_shift);

            phase += frequency;
            if (phase < srate) {
                phase += srate;
            } else if (phase > srate) {
                phase -= srate;
            }
        }
    } else {
        for (s = 0; s < sample_count; s++) {
            buffer_write(output[s], 0.0f);
        }
    }

    plugin->phase = phase;
}

void _init()
{
    char **port_names;
    LADSPA_PortDescriptor *port_descriptors;
    LADSPA_PortRangeHint *port_range_hints;
    LADSPA_Descriptor *descriptor;
    int i;

    const char *labels[] = {"pulse_fapa_oa",
                            "pulse_fapc_oa",
                            "pulse_fcpa_oa",
                            "pulse_fcpc_oa"};
    LADSPA_PortDescriptor frequency_port_descriptors[] = {LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO,
                                                          LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO,
                                                          LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL,
                                                          LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL};
    LADSPA_PortDescriptor pulsewidth_port_descriptors[] = {LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO,
                                                           LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL,
                                                           LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO,
                                                           LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL};
    LADSPA_PortDescriptor output_port_descriptors[] = {LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
                                                       LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
                                                       LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
                                                       LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO};

    void (*run_functions[])(LADSPA_Handle,
                            unsigned long) = {runPulse_fapa_oa,
                                              runPulse_fapc_oa,
                                              runPulse_fcpa_oa,
                                              runPulse_fcpc_oa};

    pulse_descriptors = (LADSPA_Descriptor **)calloc(PULSE_VARIANT_COUNT, sizeof(LADSPA_Descriptor));

    if (pulse_descriptors) {
        for (i = 0; i < PULSE_VARIANT_COUNT; i++) {
            pulse_descriptors[i] = (LADSPA_Descriptor *)malloc(sizeof(LADSPA_Descriptor));
            descriptor = pulse_descriptors[i];
            if (descriptor) {
                descriptor->UniqueID = PULSE_BASE_ID + i;
                descriptor->Label = strdup(labels[i]);
                descriptor->Properties = LADSPA_PROPERTY_HARD_RT_CAPABLE;
                descriptor->Name = strdup("Band-limited Variable Pulse Oscillator");
                descriptor->Maker = strdup("Mike Rawes <mike_rawes[at]yahoo.co.uk>");
                descriptor->Copyright = strdup("GPL");

                descriptor->PortCount = 3;

                port_descriptors = (LADSPA_PortDescriptor *)calloc(3, sizeof(LADSPA_PortDescriptor));
                descriptor->PortDescriptors = (const LADSPA_PortDescriptor *)port_descriptors;

                port_range_hints = (LADSPA_PortRangeHint *)calloc(3, sizeof(LADSPA_PortRangeHint));
                descriptor->PortRangeHints = (const LADSPA_PortRangeHint *)port_range_hints;

                port_names = (char **)calloc(3, sizeof(char*));
                descriptor->PortNames = (const char **)port_names;

                /* Parameters for Frequency */
                port_descriptors[PULSE_FREQUENCY] = frequency_port_descriptors[i];
                port_names[PULSE_FREQUENCY] = strdup("Frequency");
                port_range_hints[PULSE_FREQUENCY].HintDescriptor = LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_SAMPLE_RATE | LADSPA_HINT_LOGARITHMIC;
                port_range_hints[PULSE_FREQUENCY].LowerBound = 0.0f;
                port_range_hints[PULSE_FREQUENCY].UpperBound = 0.5f;

                /* Parameters for Pulse Width */
                port_descriptors[PULSE_PULSEWIDTH] = pulsewidth_port_descriptors[i];
                port_names[PULSE_PULSEWIDTH] = strdup("Pulse Width");
                port_range_hints[PULSE_PULSEWIDTH].HintDescriptor = LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
                port_range_hints[PULSE_PULSEWIDTH].LowerBound = 0.0f;
                port_range_hints[PULSE_PULSEWIDTH].UpperBound = 1.0f;

                /* Parameters for Output */
                port_descriptors[PULSE_OUTPUT] = output_port_descriptors[i];
                port_names[PULSE_OUTPUT] = strdup("Output");
                port_range_hints[PULSE_OUTPUT].HintDescriptor = 0;

                descriptor->activate = NULL;
                descriptor->cleanup = cleanupPulse;
                descriptor->connect_port = connectPortPulse;
                descriptor->deactivate = NULL;
                descriptor->instantiate = instantiatePulse;
                descriptor->run = run_functions[i];
                descriptor->run_adding = NULL;
                descriptor->set_run_adding_gain = NULL;
            }
        }
    }
}

void _fini() 
{
    LADSPA_Descriptor *descriptor;
    int i, j;

    if (pulse_descriptors) {
        for (i = 0; i < PULSE_VARIANT_COUNT; i++) {
            descriptor = pulse_descriptors[i];
            if (descriptor) {
                free((char *)descriptor->Label);
                free((char *)descriptor->Name);
                free((char *)descriptor->Maker);
                free((char *)descriptor->Copyright);
                free((LADSPA_PortDescriptor *)descriptor->PortDescriptors);

                for (j = 0; j < descriptor->PortCount; j++) {
                    free((char *)(descriptor->PortNames[j]));
                }
                free((char **)descriptor->PortNames);
                free((LADSPA_PortRangeHint *)descriptor->PortRangeHints);
                free(descriptor);
            }
        }
        free(pulse_descriptors);
    }
}
