/*
    random.so.c - A LADSPA plugin to generate a 'random' wave of varying
                  frequency and smoothness

    Copyright (C) 2002  Mike Rawes

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdlib.h>
#include <string.h>
#include <ladspa.h>
#include <math.h>
#include <time.h>

#define RANDOM_BASE_ID                 1661
#define RANDOM_VARIANT_COUNT           4

#define RANDOM_FREQUENCY               0
#define RANDOM_SMOOTH                  1
#define RANDOM_OUTPUT                  2

LADSPA_Descriptor **random_descriptors = NULL;

typedef struct {
    LADSPA_Data *frequency;
    LADSPA_Data *smooth;
    LADSPA_Data *output;
    LADSPA_Data srate;
    LADSPA_Data phase;
    LADSPA_Data value1;
    LADSPA_Data value2;
} Random;

LADSPA_Data inv_rand_max;

const LADSPA_Descriptor *ladspa_descriptor(unsigned long index)
{
    if (index < RANDOM_VARIANT_COUNT) {
        return random_descriptors[index];
    } else {
        return NULL;
    }
}

void cleanupRandom(LADSPA_Handle instance)
{
    free(instance);
}

void connectPortRandom(LADSPA_Handle instance,
                       unsigned long port,
                       LADSPA_Data *data)
{
    Random *plugin;

    plugin = (Random *)instance;
    switch (port) {
    case RANDOM_FREQUENCY:
        plugin->frequency = data;
        break;
    case RANDOM_SMOOTH:
        plugin->smooth = data;
        break;
    case RANDOM_OUTPUT:
        plugin->output = data;
        break;
    }
}

LADSPA_Handle instantiateRandom(const LADSPA_Descriptor *descriptor,
                                unsigned long sample_rate)
{
    Random *plugin = (Random *)malloc(sizeof(Random));

    srand((int)time(NULL));

    inv_rand_max = 2.0f / (LADSPA_Data)RAND_MAX;

    plugin->srate = (LADSPA_Data)sample_rate;
    plugin->phase = 0.0f;

    plugin->value1 = rand() * inv_rand_max - 1.0f;
    plugin->value2 = rand() * inv_rand_max - 1.0f;

    return (LADSPA_Handle)plugin;
}

#undef buffer_write
#undef RUN_ADDING
#undef RUN_REPLACING

#define buffer_write(b, v) (b = v)
#define RUN_ADDING    0
#define RUN_REPLACING 1

void runRandom_fasa_oa(LADSPA_Handle instance,
                       unsigned long sample_count)
{
    Random *plugin = (Random *)instance;

    /* Frequency (Hz) (array of floats of length sample_count) */
    LADSPA_Data *frequency = plugin->frequency;

    /* Wave smoothness (array of floats of length sample_count) */
    LADSPA_Data *smooth = plugin->smooth;

    /* Output (pointer to float value) */
    LADSPA_Data *output = plugin->output;

    /* Instance data */
    LADSPA_Data srate = plugin->srate;
    LADSPA_Data phase = plugin->phase;
    LADSPA_Data value1 = plugin->value1;
    LADSPA_Data value2 = plugin->value2;

    LADSPA_Data freq;
    LADSPA_Data smth;
    LADSPA_Data interval;
    LADSPA_Data result;
    unsigned int s;

    for (s = 0; s < sample_count; s++) {
        freq = frequency[s];

        smth = smooth[s];
        interval = (1.0f - smth) * 0.5f;

        if (phase < interval) {
            result = 1.0f;
        } else if (phase > (1.0f - interval)) {
            result = -1.0f;
        } else {
            if (interval > 0.0f) {
                result = cos((phase - interval) / smth * M_PI);
            } else {
                result = cos(phase * M_PI);
            }
        }

        result *= (value2 - value1) * 0.5f;
        result -= (value2 + value1) * 0.5f;

        buffer_write(output[s], result);

        phase += 2.0f * freq / srate;
        if (phase > 1.0f) {
            phase -= 1.0f;
            value1 = value2;
            value2 = (LADSPA_Data)rand() * inv_rand_max - 1.0f;
        }
    }

    plugin->phase = phase;
    plugin->value1 = value1;
    plugin->value2 = value2;
}

void runRandom_fasc_oa(LADSPA_Handle instance,
                       unsigned long sample_count)
{
    Random *plugin = (Random *)instance;

    /* Frequency (Hz) (array of floats of length sample_count) */
    LADSPA_Data *frequency = plugin->frequency;

    /* Wave smoothness (array of floats of length sample_count) */
    LADSPA_Data smooth = *(plugin->smooth);

    /* Output (pointer to float value) */
    LADSPA_Data *output = plugin->output;

    /* Instance data */
    LADSPA_Data srate = plugin->srate;
    LADSPA_Data phase = plugin->phase;
    LADSPA_Data value1 = plugin->value1;
    LADSPA_Data value2 = plugin->value2;

    LADSPA_Data freq;
    LADSPA_Data interval = (1.0f - smooth) * 0.5f;
    LADSPA_Data result;
    unsigned int s;

    for (s = 0; s < sample_count; s++) {
        freq = frequency[s];

        if (phase < interval) {
            result = 1.0f;
        } else if (phase > (1.0f - interval)) {
            result = -1.0f;
        } else {
            if (interval > 0.0f) {
                result = cos((phase - interval) / smooth * M_PI);
            } else {
                result = cos(phase * M_PI);
            }
        }

        result *= (value2 - value1) * 0.5f;
        result -= (value2 + value1) * 0.5f;

        buffer_write(output[s], result);

        phase += 2.0f * freq / srate;
        if (phase > 1.0f) {
            phase -= 1.0f;
            value1 = value2;
            value2 = (LADSPA_Data)rand() * inv_rand_max - 1.0f;
        }
    }

    plugin->phase = phase;
    plugin->value1 = value1;
    plugin->value2 = value2;
}

void runRandom_fcsa_oa(LADSPA_Handle instance,
                       unsigned long sample_count)
{
    Random *plugin = (Random *)instance;

    /* Frequency (Hz) (array of floats of length sample_count) */
    LADSPA_Data frequency = *(plugin->frequency);

    /* Wave smoothness (array of floats of length sample_count) */
    LADSPA_Data *smooth = plugin->smooth;

    /* Output (pointer to float value) */
    LADSPA_Data *output = plugin->output;

    /* Instance data */
    LADSPA_Data srate = plugin->srate;
    LADSPA_Data phase = plugin->phase;
    LADSPA_Data value1 = plugin->value1;
    LADSPA_Data value2 = plugin->value2;

    LADSPA_Data phase_scale = 2.0f * frequency / srate;
    LADSPA_Data smth;
    LADSPA_Data interval;
    LADSPA_Data result;
    unsigned int s;

    for (s = 0; s < sample_count; s++) {
        smth = smooth[s];
        interval = (1.0f - smth) * 0.5f;

        if (phase < interval) {
            result = 1.0f;
        } else if (phase > (1.0f - interval)) {
            result = -1.0f;
        } else {
            if (interval > 0.0f) {
                result = cos((phase - interval) / smth * M_PI);
            } else {
                result = cos(phase * M_PI);
            }
        }

        result *= (value2 - value1) * 0.5f;
        result -= (value2 + value1) * 0.5f;

        buffer_write(output[s], result);

        phase += phase_scale;
        if (phase > 1.0f) {
            phase -= 1.0f;
            value1 = value2;
            value2 = (LADSPA_Data)rand() * inv_rand_max - 1.0f;
        }
    }

    plugin->phase = phase;
    plugin->value1 = value1;
    plugin->value2 = value2;
}

void runRandom_fcsc_oa(LADSPA_Handle instance,
                       unsigned long sample_count)
{
    Random *plugin = (Random *)instance;

    /* Frequency (Hz) (array of floats of length sample_count) */
    LADSPA_Data frequency = *(plugin->frequency);

    /* Wave smoothness (array of floats of length sample_count) */
    LADSPA_Data smooth = *(plugin->smooth);

    /* Output (pointer to float value) */
    LADSPA_Data *output = plugin->output;

    /* Instance data */
    LADSPA_Data srate = plugin->srate;
    LADSPA_Data phase = plugin->phase;
    LADSPA_Data value1 = plugin->value1;
    LADSPA_Data value2 = plugin->value2;

    LADSPA_Data phase_scale = 2.0f * frequency / srate;
    LADSPA_Data interval = (1.0f - smooth) * 0.5f;
    LADSPA_Data result;
    unsigned int s;

    for (s = 0; s < sample_count; s++) {
        if (phase < interval) {
            result = 1.0f;
        } else if (phase > (1.0f - interval)) {
            result = -1.0f;
        } else {
            if (interval > 0.0f) {
                result = cos((phase - interval) / smooth * M_PI);
            } else {
                result = cos(phase * M_PI);
            }
        }

        result *= (value2 - value1) * 0.5f;
        result -= (value2 + value1) * 0.5f;

        buffer_write(output[s], result);

        phase += phase_scale;
        if (phase > 1.0f) {
            phase -= 1.0f;
            value1 = value2;
            value2 = (LADSPA_Data)rand() * inv_rand_max - 1.0f;
        }
    }

    plugin->phase = phase;
    plugin->value1 = value1;
    plugin->value2 = value2;
}

void _init()
{
    char **port_names;
    LADSPA_PortDescriptor *port_descriptors;
    LADSPA_PortRangeHint *port_range_hints;
    LADSPA_Descriptor *descriptor;
    int i;

    const char *labels[] = {"random_fasa_oa",
                            "random_fasc_oa",
                            "random_fcsa_oa",
                            "random_fcsc_oa"};
    LADSPA_PortDescriptor frequency_port_descriptors[] = {LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO,
                                                          LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO,
                                                          LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL,
                                                          LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL};
    LADSPA_PortDescriptor smooth_port_descriptors[] = {LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO,
                                                       LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL,
                                                       LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO,
                                                       LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL};
    LADSPA_PortDescriptor output_port_descriptors[] = {LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
                                                       LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
                                                       LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
                                                       LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO};

    void (*run_functions[])(LADSPA_Handle,
                            unsigned long) = {runRandom_fasa_oa,
                                              runRandom_fasc_oa,
                                              runRandom_fcsa_oa,
                                              runRandom_fcsc_oa};

    random_descriptors = (LADSPA_Descriptor **)calloc(RANDOM_VARIANT_COUNT, sizeof(LADSPA_Descriptor));

    if (random_descriptors) {
        for (i = 0; i < RANDOM_VARIANT_COUNT; i++) {
            random_descriptors[i] = (LADSPA_Descriptor *)malloc(sizeof(LADSPA_Descriptor));
            descriptor = random_descriptors[i];
            if (descriptor) {
                descriptor->UniqueID = RANDOM_BASE_ID + i;
                descriptor->Label = strdup(labels[i]);
                descriptor->Properties = LADSPA_PROPERTY_HARD_RT_CAPABLE;
                descriptor->Name = strdup("Random Wave");
                descriptor->Maker = strdup("Mike Rawes <mike_rawes[at]yahoo.co.uk>");
                descriptor->Copyright = strdup("GPL");

                descriptor->PortCount = 3;

                port_descriptors = (LADSPA_PortDescriptor *)calloc(3, sizeof(LADSPA_PortDescriptor));
                descriptor->PortDescriptors = (const LADSPA_PortDescriptor *)port_descriptors;

                port_range_hints = (LADSPA_PortRangeHint *)calloc(3, sizeof(LADSPA_PortRangeHint));
                descriptor->PortRangeHints = (const LADSPA_PortRangeHint *)port_range_hints;

                port_names = (char **)calloc(3, sizeof(char*));
                descriptor->PortNames = (const char **)port_names;

                /* Parameters for Frequency (Hz) */
                port_descriptors[RANDOM_FREQUENCY] = frequency_port_descriptors[i];
                port_names[RANDOM_FREQUENCY] = strdup("Frequency (Hz)");
                port_range_hints[RANDOM_FREQUENCY].HintDescriptor = LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_SAMPLE_RATE | LADSPA_HINT_LOGARITHMIC;
                port_range_hints[RANDOM_FREQUENCY].LowerBound = 0.0f;
                port_range_hints[RANDOM_FREQUENCY].UpperBound = 0.5f;

                /* Parameters for Wave smoothness */
                port_descriptors[RANDOM_SMOOTH] = smooth_port_descriptors[i];
                port_names[RANDOM_SMOOTH] = strdup("Wave Smoothness");
                port_range_hints[RANDOM_SMOOTH].HintDescriptor = LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
                port_range_hints[RANDOM_SMOOTH].LowerBound = 0.0f;
                port_range_hints[RANDOM_SMOOTH].UpperBound = 1.0f;

                /* Parameters for Output */
                port_descriptors[RANDOM_OUTPUT] = output_port_descriptors[i];
                port_names[RANDOM_OUTPUT] = strdup("Output");
                port_range_hints[RANDOM_OUTPUT].HintDescriptor = 0;

                descriptor->activate = NULL;
                descriptor->cleanup = cleanupRandom;
                descriptor->connect_port = connectPortRandom;
                descriptor->deactivate = NULL;
                descriptor->instantiate = instantiateRandom;
                descriptor->run = run_functions[i];
                descriptor->run_adding = NULL;
                descriptor->set_run_adding_gain = NULL;
            }
        }
    }
}

void _fini()
{
    LADSPA_Descriptor *descriptor;
    int i, j;

    if (random_descriptors) {
        for (i = 0; i < RANDOM_VARIANT_COUNT; i++) {
            descriptor = random_descriptors[i];
            if (descriptor) {
                free((char *)descriptor->Label);
                free((char *)descriptor->Name);
                free((char *)descriptor->Maker);
                free((char *)descriptor->Copyright);
                free((LADSPA_PortDescriptor *)descriptor->PortDescriptors);

                for (j = 0; j < descriptor->PortCount; j++) {
                    free((char *)(descriptor->PortNames[j]));
                }
                free((char **)descriptor->PortNames);
                free((LADSPA_PortRangeHint *)descriptor->PortRangeHints);
                free(descriptor);
            }
        }
        free(random_descriptors);
    }
}
