/*
    sawtooth.so.c - A LADSPA plugin to generate a bandlimited sawtooth waveform

    Copyright (C) 2002  Mike Rawes

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "ladspa.h"
#include "interpolate.h"
#include "wavedata.h"

/* This contains all the generated table data */
#include "sawtooth_data.h"

#define INTERPOLATE(i,a,b,c,d) interpolateCubic(i,a,b,c,d)

#define SAWTOOTH_BASE_ID            1641
#define SAWTOOTH_VARIANT_COUNT      2

#define SAWTOOTH_FREQUENCY          0
#define SAWTOOTH_OUTPUT             1

LADSPA_Descriptor **sawtooth_descriptors = NULL;

typedef struct {
    LADSPA_Data *frequency;
    LADSPA_Data *output;
    Wavedata *wdat;
    LADSPA_Data srate;
    LADSPA_Data phase;
} Sawtooth;

/*****************************************************************************
 *
 * LADSPA Plugin code
 *
 *****************************************************************************/

const LADSPA_Descriptor *ladspa_descriptor(unsigned long index)
{
    if (index < 2) {
        return sawtooth_descriptors[index];
    } else {
        return NULL;
    }
}

void cleanupSawtooth(LADSPA_Handle instance)
{
    Sawtooth *plugin = (Sawtooth *)instance;

    destroyWavedata(plugin->wdat);
    free(instance);
}

void connectPortSawtooth(LADSPA_Handle instance,
                         unsigned long port,
                         LADSPA_Data *data)
{
    Sawtooth *plugin;

    plugin = (Sawtooth *)instance;
    switch (port) {
    case SAWTOOTH_FREQUENCY:
        plugin->frequency = data;
        break;
    case SAWTOOTH_OUTPUT:
        plugin->output = data;
        break;
    }
}

LADSPA_Handle instantiateSawtooth(const LADSPA_Descriptor *descriptor,
                                  unsigned long sample_rate)
{
    Sawtooth *plugin = (Sawtooth *)malloc(sizeof(Sawtooth));

    Wdatheader saw_data;

/* Pass data from header to wavedata initialisation */
    saw_data.number_of_wavetables = wdat_number_of_wavetables;
    saw_data.lowest_harmonic = wdat_lowest_harmonic;
    saw_data.highest_harmonic = wdat_highest_harmonic;
    saw_data.max_harmonics = wdat_max_harmonics;
    saw_data.sample_counts = wdat_sample_counts;
    saw_data.sample_data_base = wdat_sample_data_base;
    saw_data.sample_data_extra = wdat_sample_data_extra;

    plugin->wdat = initWavedata(&saw_data, sample_rate);
    if (!plugin->wdat) {
        return NULL;
    }
    plugin->srate = plugin->wdat->sample_rate;
    plugin->phase = 0.0f;

    return (LADSPA_Handle)plugin;
}

#undef buffer_write
#undef RUN_ADDING
#undef RUN_REPLACING

#define buffer_write(b, v) (b = v)
#define RUN_ADDING    0
#define RUN_REPLACING 1

void runSawtooth_fa_oa(LADSPA_Handle instance,
                       unsigned long sample_count)
{
    Sawtooth *plugin = (Sawtooth *)instance;

    /* Frequency (array of LADSPA_Data of length sample_count) */
    LADSPA_Data *frequency = plugin->frequency;

    /* Output (pointer to LADSPA_Data value) */
    LADSPA_Data *output = plugin->output;

    /* Instance data */
    Wavedata *wdat = plugin->wdat;
    LADSPA_Data phase = plugin->phase;
    LADSPA_Data srate = plugin->srate;

    Wavetable *wtbl;
    LADSPA_Data freq;
    LADSPA_Data abs_freq;
    LADSPA_Data interval;
    LADSPA_Data p1, p0, n0, n1;
    LADSPA_Data sample;
    unsigned long s;

    for (s = 0; s < sample_count; s++) {
        freq = frequency[s];
        abs_freq = (LADSPA_Data)fabsf((float)freq);

    /* See which table to play */
        wtbl = getTable(wdat, abs_freq);
        if (wtbl) {
        /* Get interpolated sample from wavetable */
            interval = getSamples(wtbl, abs_freq, phase, &p1, &p0, &n0, &n1);
            sample = INTERPOLATE(interval, p1, p0, n0, n1);

            buffer_write(output[s], sample);

        /* Update phase, wrapping if necessary */
            phase += freq;
            if (phase < srate) {
                phase += srate;
            } else if (phase > srate) {
                phase -= srate;
            }
        } else {
        /* No table found (inplies frequency exceeds nyquist) */
            buffer_write(output[s], 0.0f);
        }
    }

    plugin->phase = phase;
}

void runSawtooth_fc_oa(LADSPA_Handle instance,
                       unsigned long sample_count)
{
    Sawtooth *plugin = (Sawtooth *)instance;

    /* Frequency (LADSPA_Data value) */
    LADSPA_Data frequency = *(plugin->frequency);

    /* Output (pointer to LADSPA_Data value) */
    LADSPA_Data *output = plugin->output;

    /* Instance data */
    Wavedata *wdat = plugin->wdat;
    LADSPA_Data phase = plugin->phase;
    LADSPA_Data srate = plugin->srate;

    Wavetable *wtbl;
    LADSPA_Data abs_freq = (LADSPA_Data)fabsf((float)frequency);
    LADSPA_Data interval;
    LADSPA_Data p1, p0, n0, n1;
    LADSPA_Data sample;
    unsigned long s;

    wtbl = getTable(wdat, abs_freq);
    if (wtbl) {
        for (s = 0; s < sample_count; s++) {
            interval = getSamples(wtbl, abs_freq, phase, &p1, &p0, &n0, &n1);
            sample = INTERPOLATE(interval, p1, p0, n0, n1);

            buffer_write(output[s], sample);

            phase += frequency;
            if (phase < srate) {
                phase += srate;
            } else if (phase > srate) {
                phase -= srate;
            }
        }
    } else {
    /* No table found (inplies frequency exceeds nyquist) */
        for (s = 0; s < sample_count; s++) {
            buffer_write(output[s], 0.0f);
        }
    }

    plugin->phase = phase;
}

void _init()
{
    char **port_names;
    LADSPA_PortDescriptor *port_descriptors;
    LADSPA_PortRangeHint *port_range_hints;
    LADSPA_Descriptor *descriptor;
    int i;

    const char *labels[] = {"sawtooth_fa_oa", "sawtooth_fc_oa"};
    LADSPA_PortDescriptor frequency_port_descriptors[] = {LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO,
                                                          LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL};
    LADSPA_PortDescriptor output_port_descriptors[] = {LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
                                                       LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO};

    void (*run_functions[])(LADSPA_Handle,
                            unsigned long) = {runSawtooth_fa_oa,
                                              runSawtooth_fc_oa};

    sawtooth_descriptors = (LADSPA_Descriptor **)calloc(SAWTOOTH_VARIANT_COUNT, sizeof(LADSPA_Descriptor));

    if (sawtooth_descriptors) {
        for (i = 0; i < SAWTOOTH_VARIANT_COUNT; i++) {
            sawtooth_descriptors[i] = (LADSPA_Descriptor *)malloc(sizeof(LADSPA_Descriptor));
            descriptor = sawtooth_descriptors[i];
            if (descriptor) {
                descriptor->UniqueID = SAWTOOTH_BASE_ID + i;
                descriptor->Label = strdup(labels[i]);
                descriptor->Properties = LADSPA_PROPERTY_HARD_RT_CAPABLE;
                descriptor->Name = strdup("Bandlimited Sawtooth Oscillator");
                descriptor->Maker = strdup("Mike Rawes <mike_rawes[at]yahoo.co.uk>");
                descriptor->Copyright = strdup("GPL");

                descriptor->PortCount = 2;

                port_descriptors = (LADSPA_PortDescriptor *)calloc(2, sizeof(LADSPA_PortDescriptor));
                descriptor->PortDescriptors = (const LADSPA_PortDescriptor *)port_descriptors;

                port_range_hints = (LADSPA_PortRangeHint *)calloc(2, sizeof(LADSPA_PortRangeHint));
                descriptor->PortRangeHints = (const LADSPA_PortRangeHint *)port_range_hints;

                port_names = (char **)calloc(2, sizeof(char*));
                descriptor->PortNames = (const char **)port_names;

                /* Parameters for Frequency */
                port_descriptors[SAWTOOTH_FREQUENCY] = frequency_port_descriptors[i];
                port_names[SAWTOOTH_FREQUENCY] = strdup("Frequency");
                port_range_hints[SAWTOOTH_FREQUENCY].HintDescriptor = LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_SAMPLE_RATE | LADSPA_HINT_LOGARITHMIC;
                port_range_hints[SAWTOOTH_FREQUENCY].LowerBound = 0.0f;
                port_range_hints[SAWTOOTH_FREQUENCY].UpperBound = 0.5f;

                /* Parameters for Output */
                port_descriptors[SAWTOOTH_OUTPUT] = output_port_descriptors[i];
                port_names[SAWTOOTH_OUTPUT] = strdup("Output");
                port_range_hints[SAWTOOTH_OUTPUT].HintDescriptor = 0;

                descriptor->activate = NULL;
                descriptor->cleanup = cleanupSawtooth;
                descriptor->connect_port = connectPortSawtooth;
                descriptor->deactivate = NULL;
                descriptor->instantiate = instantiateSawtooth;
                descriptor->run = run_functions[i];
                descriptor->run_adding = NULL;
                descriptor->set_run_adding_gain = NULL;
            }
        }
    }
}

void _fini() 
{
    LADSPA_Descriptor *descriptor;
    int i, j;

    if (sawtooth_descriptors) {
        for (i = 0; i < SAWTOOTH_VARIANT_COUNT; i++) {
            descriptor = sawtooth_descriptors[i];
            if (descriptor) {
                free((char *)descriptor->Label);
                free((char *)descriptor->Name);
                free((char *)descriptor->Maker);
                free((char *)descriptor->Copyright);
                free((LADSPA_PortDescriptor *)descriptor->PortDescriptors);

                for (j = 0; j < descriptor->PortCount; j++) {
                    free((char *)(descriptor->PortNames[j]));
                }
                free((char **)descriptor->PortNames);
                free((LADSPA_PortRangeHint *)descriptor->PortRangeHints);
                free(descriptor);
            }
        }
        free(sawtooth_descriptors);
    }
}
