/*
    triangle.so.c - A LADSPA plugin to generate a bandlimited slope-variable
                    triangle waveform

    Copyright (C) 2002  Mike Rawes

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "ladspa.h"
#include "interpolate.h"
#include "wavedata.h"

/* This contains all the generated table data */
#include "parabola_data.h"

#define INTERPOLATE(i,a,b,c,d) interpolateCubic(i,a,b,c,d)

#define TRIANGLE_BASE_ID               1649
#define TRIANGLE_VARIANT_COUNT         4

#define TRIANGLE_FREQUENCY             0
#define TRIANGLE_SLOPE                 1
#define TRIANGLE_OUTPUT                2

LADSPA_Descriptor **triangle_descriptors = NULL;

typedef struct {
    LADSPA_Data *frequency;
    LADSPA_Data *slope;
    LADSPA_Data *output;
    Wavedata *wdat;
    LADSPA_Data srate;
    LADSPA_Data phase;
    LADSPA_Data min_slope;
    LADSPA_Data max_slope;
} Triangle;

Wavetable *wtbl;
LADSPA_Data freq;
LADSPA_Data abs_freq;
LADSPA_Data slp;
LADSPA_Data shifted_phase;
LADSPA_Data scale;
LADSPA_Data interval;
LADSPA_Data p1, p0, n0, n1;
LADSPA_Data sample;

/*****************************************************************************
 *
 * LADSPA Plugin code
 *
 *****************************************************************************/

const LADSPA_Descriptor *ladspa_descriptor(unsigned long index)
{
    if (index < TRIANGLE_VARIANT_COUNT) {
        return triangle_descriptors[index];
    } else {
        return NULL;
    }
}

void cleanupTriangle(LADSPA_Handle instance)
{
    Triangle *plugin = (Triangle *)instance;

    destroyWavedata(plugin->wdat);
    free(instance);
}

void connectPortTriangle(LADSPA_Handle instance,
                         unsigned long port,
                         LADSPA_Data *data)
{
    Triangle *plugin = (Triangle *)instance;

    switch (port) {
    case TRIANGLE_FREQUENCY:
        plugin->frequency = data;
        break;
    case TRIANGLE_SLOPE:
        plugin->slope = data;
        break;
    case TRIANGLE_OUTPUT:
        plugin->output = data;
        break;
    }
}

LADSPA_Handle instantiateTriangle(const LADSPA_Descriptor *descriptor,
                                  unsigned long sample_rate)
{
    Triangle *plugin = (Triangle *)malloc(sizeof(Triangle));

    Wdatheader prb_data;

/* Pass data from header to wavedata initialisation */
    prb_data.number_of_wavetables = wdat_number_of_wavetables;
    prb_data.lowest_harmonic = wdat_lowest_harmonic;
    prb_data.highest_harmonic = wdat_highest_harmonic;
    prb_data.max_harmonics = wdat_max_harmonics;
    prb_data.sample_counts = wdat_sample_counts;
    prb_data.sample_data_base = wdat_sample_data_base;
    prb_data.sample_data_extra = wdat_sample_data_extra;

    plugin->wdat = initWavedata(&prb_data, sample_rate);
    if (!plugin->wdat) {
        return NULL;
    }
    plugin->srate = plugin->wdat->sample_rate;
    plugin->phase = 0.0f;

    plugin->min_slope = 2.0f / plugin->srate;
    plugin->max_slope = 1.0f - plugin->min_slope;

    return (LADSPA_Handle)plugin;
}

#undef buffer_write
#undef RUN_ADDING
#undef RUN_REPLACING

#define buffer_write(b, v) (b = v)
#define RUN_ADDING    0
#define RUN_REPLACING 1

void runTriangle_fasa_oa(LADSPA_Handle instance,
                         unsigned long sample_count)
{
    Triangle *plugin = (Triangle *)instance;

    /* Frequency (array of LADSPA_Data of length sample_count) */
    LADSPA_Data *frequency = plugin->frequency;

    /* Slope (array of LADSPA_Data of length sample_count) */
    LADSPA_Data *slope = plugin->slope;

    /* Output (pointer to LADSPA_Data value) */
    LADSPA_Data *output = plugin->output;

    /* Instance data */
    Wavedata *wdat = plugin->wdat;
    LADSPA_Data srate = plugin->srate;
    LADSPA_Data phase = plugin->phase;
    LADSPA_Data min_slope = plugin->min_slope;
    LADSPA_Data max_slope = plugin->max_slope;

    Wavetable *wtbl;
    LADSPA_Data freq;
    LADSPA_Data abs_freq;
    LADSPA_Data slp;
    LADSPA_Data shifted_phase;
    LADSPA_Data interval;
    LADSPA_Data p1, p0, n0, n1;
    LADSPA_Data sample;
    unsigned long s;

    for (s = 0; s < sample_count; s++) {
        freq = frequency[s];
        abs_freq = (LADSPA_Data)fabsf((float)freq);
        slp = slope[s];

    /* Lookup which table to use from frequency */
        wtbl = getTable(wdat, abs_freq);
        if (wtbl) {
        /* Get samples from parabola and phase shifted inverted parabola */
            interval = getSamples(wtbl, abs_freq, phase, &p1, &p0, &n0, &n1);
            sample = INTERPOLATE(interval, p1, p0, n0, n1);

            slp = slp < min_slope ? min_slope : slp;
            slp = slp > max_slope ? max_slope : slp;
            shifted_phase = phase + slp * srate;

            interval = getSamples(wtbl, abs_freq, shifted_phase, &p1, &p0, &n0, &n1);
            sample -= INTERPOLATE(interval, p1, p0, n0, n1);

        /* Scale result due to phase difference */
            buffer_write(output[s], sample / (8.0f * (slp - (slp * slp))));

            phase += freq;
            if (phase < srate) {
                phase += srate;
            } else if (phase > srate) {
                phase -= srate;
            }
        } else {
        /* No table found (inplies frequency exceeds nyquist) */
            buffer_write(output[s], 0.0f);
        }
    }

    plugin->phase = phase;
}

void runTriangle_fasc_oa(LADSPA_Handle instance,
                         unsigned long sample_count)
{
    Triangle *plugin = (Triangle *)instance;

    /* Frequency (array of LADSPA_Data of length sample_count) */
    LADSPA_Data *frequency = plugin->frequency;

    /* Slope (LADSPA_Data value) */
    LADSPA_Data slope = *(plugin->slope);

    /* Output (pointer to LADSPA_Data value) */
    LADSPA_Data *output = plugin->output;

    /* Instance data */
    Wavedata *wdat = plugin->wdat;
    LADSPA_Data srate = plugin->srate;
    LADSPA_Data phase = plugin->phase;
    LADSPA_Data min_slope = plugin->min_slope;
    LADSPA_Data max_slope = plugin->max_slope;

    Wavetable *wtbl;
    LADSPA_Data freq;
    LADSPA_Data abs_freq;
    LADSPA_Data shifted_phase;
    LADSPA_Data scale;
    LADSPA_Data interval;
    LADSPA_Data p1, p0, n0, n1;
    LADSPA_Data sample;
    unsigned long s;

    slope = slope < min_slope ? min_slope : slope;
    slope = slope > max_slope ? max_slope : slope;
    scale = 1.0f / (8.0f * (slope - (slope * slope)));

    for (s = 0; s < sample_count; s++) {
        freq = frequency[s];
        abs_freq = (LADSPA_Data)fabsf((float)freq);

    /* Lookup which table to use from frequency */
        wtbl = getTable(wdat, abs_freq);
        if (wtbl) {
        /* Get samples from parabola and phase shifted inverted parabola */
            interval = getSamples(wtbl, abs_freq, phase, &p1, &p0, &n0, &n1);
            sample = INTERPOLATE(interval, p1, p0, n0, n1);

            shifted_phase = phase + slope * srate;

            interval = getSamples(wtbl, abs_freq, shifted_phase, &p1, &p0, &n0, &n1);
            sample -= INTERPOLATE(interval, p1, p0, n0, n1);

        /* Scale result due to phase difference */
            buffer_write(output[s], sample * scale);

            phase += freq;
            if (phase < srate) {
                phase += srate;
            } else if (phase > srate) {
                phase -= srate;
            }
        } else {
        /* No table found (inplies frequency exceeds nyquist) */
            buffer_write(output[s], 0.0f);
        }
    }

    plugin->phase = phase;
}

void runTriangle_fcsa_oa(LADSPA_Handle instance,
                         unsigned long sample_count)
{
    Triangle *plugin = (Triangle *)instance;

    /* Frequency (LADSPA_Data value) */
    LADSPA_Data frequency = *(plugin->frequency);

    /* Slope (array of LADSPA_Data of length sample_count) */
    LADSPA_Data *slope = plugin->slope;

    /* Output (pointer to LADSPA_Data value) */
    LADSPA_Data *output = plugin->output;

    /* Instance data */
    Wavedata *wdat = plugin->wdat;
    LADSPA_Data srate = plugin->srate;
    LADSPA_Data phase = plugin->phase;
    LADSPA_Data min_slope = plugin->min_slope;
    LADSPA_Data max_slope = plugin->max_slope;

    Wavetable *wtbl;
    LADSPA_Data abs_freq = (LADSPA_Data)fabsf((float)frequency);
    LADSPA_Data slp;
    LADSPA_Data shifted_phase;
    LADSPA_Data interval;
    LADSPA_Data p1, p0, n0, n1;
    LADSPA_Data sample;
    unsigned long s;

    wtbl = getTable(wdat, abs_freq);
    if (wtbl) {
        for (s = 0; s < sample_count; s++) {
            slp = slope[s];

        /* Get samples from parabola and phase shifted inverted parabola */
            interval = getSamples(wtbl, abs_freq, phase, &p1, &p0, &n0, &n1);
            sample = INTERPOLATE(interval, p1, p0, n0, n1);

            slp = slp < min_slope ? min_slope : slp;
            slp = slp > max_slope ? max_slope : slp;
            shifted_phase = phase + slp * srate;

            interval = getSamples(wtbl, abs_freq, shifted_phase, &p1, &p0, &n0, &n1);
            sample -= INTERPOLATE(interval, p1, p0, n0, n1);

            buffer_write(output[s], sample / (8.0f * (slp - (slp * slp))));

            phase += frequency;
            if (phase < srate) {
                phase += srate;
            } else if (phase > srate) {
                phase -= srate;
            }
        }
    } else {
        for (s = 0; s < sample_count; s++) {
        /* No table found (inplies frequency exceeds nyquist) */
            buffer_write(output[s], 0.0f);
        }
    }

    plugin->phase = phase;
}

void runTriangle_fcsc_oa(LADSPA_Handle instance,
                         unsigned long sample_count)
{
    Triangle *plugin = (Triangle *)instance;

    /* Frequency (LADSPA_Data value) */
    LADSPA_Data frequency = *(plugin->frequency);

    /* Slope (LADSPA_Data value) */
    LADSPA_Data slope = *(plugin->slope);

    /* Output (pointer to LADSPA_Data value) */
    LADSPA_Data *output = plugin->output;

    /* Instance data */
    Wavedata *wdat = plugin->wdat;
    LADSPA_Data srate = plugin->srate;
    LADSPA_Data phase = plugin->phase;
    LADSPA_Data min_slope = plugin->min_slope;
    LADSPA_Data max_slope = plugin->max_slope;

    Wavetable *wtbl;
    LADSPA_Data abs_freq = (LADSPA_Data)fabsf((float)frequency);
    LADSPA_Data shifted_phase;
    LADSPA_Data scale;
    LADSPA_Data interval;
    LADSPA_Data p1, p0, n0, n1;
    LADSPA_Data sample;
    unsigned long s;

    slope = slope < min_slope ? min_slope : slope;
    slope = slope > max_slope ? max_slope : slope;
    scale = 1.0f / (8.0f * (slope - (slope * slope)));

    wtbl = getTable(wdat, abs_freq);
    if (wtbl) {
        for (s = 0; s < sample_count; s++) {
        /* Get samples from parabola and phase shifted inverted parabola */
            interval = getSamples(wtbl, abs_freq, phase, &p1, &p0, &n0, &n1);
            sample = INTERPOLATE(interval, p1, p0, n0, n1);

            shifted_phase = phase + slope * srate;

            interval = getSamples(wtbl, abs_freq, shifted_phase, &p1, &p0, &n0, &n1);
            sample -= INTERPOLATE(interval, p1, p0, n0, n1);

            buffer_write(output[s], sample * scale);

            phase += frequency;
            if (phase < srate) {
                phase += srate;
            } else if (phase > srate) {
                phase -= srate;
            }
        }
    } else {
        for (s = 0; s < sample_count; s++) {
        /* No table found (inplies frequency exceeds nyquist) */
            buffer_write(output[s], 0.0f);
        }
    }

    plugin->phase = phase;
}

void _init()
{
    char **port_names;
    LADSPA_PortDescriptor *port_descriptors;
    LADSPA_PortRangeHint *port_range_hints;
    LADSPA_Descriptor *descriptor;
    int i;

    const char *labels[] = {"triangle_fasa_oa",
                            "triangle_fasc_oa",
                            "triangle_fcsa_oa",
                            "triangle_fcsc_oa"};
    LADSPA_PortDescriptor frequency_port_descriptors[] = {LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO,
                                                          LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO,
                                                          LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL,
                                                          LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL};
    LADSPA_PortDescriptor slope_port_descriptors[] = {LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO,
                                                      LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL,
                                                      LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO,
                                                      LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL};
    LADSPA_PortDescriptor output_port_descriptors[] = {LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
                                                       LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
                                                       LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
                                                       LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO};

    void (*run_functions[])(LADSPA_Handle,
                            unsigned long) = {runTriangle_fasa_oa,
                                              runTriangle_fasc_oa,
                                              runTriangle_fcsa_oa,
                                              runTriangle_fcsc_oa};

    triangle_descriptors = (LADSPA_Descriptor **)calloc(TRIANGLE_VARIANT_COUNT, sizeof(LADSPA_Descriptor));

    if (triangle_descriptors) {
        for (i = 0; i < TRIANGLE_VARIANT_COUNT; i++) {
            triangle_descriptors[i] = (LADSPA_Descriptor *)malloc(sizeof(LADSPA_Descriptor));
            descriptor = triangle_descriptors[i];
            if (descriptor) {
                descriptor->UniqueID = TRIANGLE_BASE_ID + i;
                descriptor->Label = strdup(labels[i]);
                descriptor->Properties = LADSPA_PROPERTY_HARD_RT_CAPABLE;
                descriptor->Name = strdup("Band-limited Slope Variable Triangle Oscillator");
                descriptor->Maker = strdup("Mike Rawes <mike_rawes[at]yahoo.co.uk>");
                descriptor->Copyright = strdup("GPL");

                descriptor->PortCount = 3;

                port_descriptors = (LADSPA_PortDescriptor *)calloc(3, sizeof(LADSPA_PortDescriptor));
                descriptor->PortDescriptors = (const LADSPA_PortDescriptor *)port_descriptors;

                port_range_hints = (LADSPA_PortRangeHint *)calloc(3, sizeof(LADSPA_PortRangeHint));
                descriptor->PortRangeHints = (const LADSPA_PortRangeHint *)port_range_hints;

                port_names = (char **)calloc(3, sizeof(char*));
                descriptor->PortNames = (const char **)port_names;

                /* Parameters for Frequency */
                port_descriptors[TRIANGLE_FREQUENCY] = frequency_port_descriptors[i];
                port_names[TRIANGLE_FREQUENCY] = strdup("Frequency");
                port_range_hints[TRIANGLE_FREQUENCY].HintDescriptor = LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_SAMPLE_RATE | LADSPA_HINT_LOGARITHMIC;
                port_range_hints[TRIANGLE_FREQUENCY].LowerBound = 0.0f;
                port_range_hints[TRIANGLE_FREQUENCY].UpperBound = 0.5f;

                /* Parameters for Slope */
                port_descriptors[TRIANGLE_SLOPE] = slope_port_descriptors[i];
                port_names[TRIANGLE_SLOPE] = strdup("Slope");
                port_range_hints[TRIANGLE_SLOPE].HintDescriptor = LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
                port_range_hints[TRIANGLE_SLOPE].LowerBound = 0.0f;
                port_range_hints[TRIANGLE_SLOPE].UpperBound = 1.0f;

                /* Parameters for Output */
                port_descriptors[TRIANGLE_OUTPUT] = output_port_descriptors[i];
                port_names[TRIANGLE_OUTPUT] = strdup("Output");
                port_range_hints[TRIANGLE_OUTPUT].HintDescriptor = 0;

                descriptor->activate = NULL;
                descriptor->cleanup = cleanupTriangle;
                descriptor->connect_port = connectPortTriangle;
                descriptor->deactivate = NULL;
                descriptor->instantiate = instantiateTriangle;
                descriptor->run = run_functions[i];
                descriptor->run_adding = NULL;
                descriptor->set_run_adding_gain = NULL;
            }
        }
    }
}

void _fini()
{
    LADSPA_Descriptor *descriptor;
    int i, j;

    if (triangle_descriptors) {
        for (i = 0; i < TRIANGLE_VARIANT_COUNT; i++) {
            descriptor = triangle_descriptors[i];
            if (descriptor) {
                free((char *)descriptor->Label);
                free((char *)descriptor->Name);
                free((char *)descriptor->Maker);
                free((char *)descriptor->Copyright);
                free((LADSPA_PortDescriptor *)descriptor->PortDescriptors);

                for (j = 0; j < descriptor->PortCount; j++) {
                    free((char *)(descriptor->PortNames[j]));
                }
                free((char **)descriptor->PortNames);
                free((LADSPA_PortRangeHint *)descriptor->PortRangeHints);
                free(descriptor);
            }
        }
        free(triangle_descriptors);
    }
}
