/*
    wdatutil.c - Code to generate wavedata for bandlimited waveforms.

    Copyright (C) 2002  Mike Rawes

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "ladspa.h"
#include "common.h"
#include "wdatutil.h"

#ifdef __cplusplus
extern "C" {
#endif

void generateSawtooth(Wtblfile *wtblfile,
                      float gibbs_comp);
void generateSquare(Wtblfile *wtblfile,
                    float gibbs_comp);
void generateParabola(Wtblfile *wtblfile,
                      float gibbs_comp);

#ifdef __cplusplus
} /* extern "C" { */
#endif

char *wave_names[] = {
    "Saw",
    "Square",
    "Parabola"
};

char *wave_descriptions[] = {
    "Sawtooth Wave",
    "Square Wave",
    "Parabola Wave"
};

unsigned long wave_first_harmonics[] = {
    1,
    1,
    1
};

unsigned long wave_harmonic_intervals[] = {
    1,
    2,
    1
};

void (*wave_functions[])(Wtblfile *wtblfile,
                         float gibbs_comp) = {
     generateSawtooth,
     generateSquare,
     generateParabola,
};

/*
 * Generate bandlimited sawtooth wave
 */
void
generateSawtooth(Wtblfile *wtblfile,
                 float gibbs_comp)
{
/* Set up to ignore first and last two samples */
    unsigned long sample_count = wtblfile->sample_count - 3;
    unsigned long start = 1, end = sample_count + 1;
    double inv_sample_countf = 1.0 / (double)sample_count;
    LADSPA_Data *samples_base = wtblfile->samples_base;
    LADSPA_Data *samples_extra = wtblfile->samples_extra;
    unsigned long max_hb = wtblfile->max_harmonic_base;
    unsigned long max_he = wtblfile->max_harmonic_extra;
    unsigned long fh = wave_first_harmonics[0];
    unsigned long hs = wave_harmonic_intervals[0];

    LADSPA_Data scale = 2.0f / M_PI;
    unsigned long i, h;
    double phase;
    double hf;
    double k;
    double m;

    if (gibbs_comp > 0.0f) {
    /* Compensate for Gibbs effect (overshoot at peaks) */
        k = M_PI * (double)gibbs_comp / ((double)max_he * 2.0);
        for (i = start; i < end; i++) {
            samples_base[i] = 0.0f;
            samples_extra[i] = 0.0f;
        }
    /* Generate base samples */
        for (h = fh; h <= max_hb; h += hs) {
            hf = (double)h;
            m = cos((hf - 1.0) * k);
            m *= m;
            for (i = start; i < end; i++) {
                phase = (double)i * 2.0 * M_PI * inv_sample_countf;
                samples_base[i] += (LADSPA_Data)((1.0 / hf) * sin(phase * hf) * m);
            }
        }
    /* Generate extra samples for higher harmonics */
        for (; h <= max_he; h += hs) {
            hf = (double)h;
            m = cos((hf - 1.0) * k);
            m *= m;
            for (i = start; i < end; i++) {
                phase = (double)i * 2.0 * M_PI * inv_sample_countf;
                samples_extra[i] += (LADSPA_Data)((1.0 / hf) * sin(phase * hf) * m);
            }
        }
        for (i = start; i < end; i++) {
            samples_base[i] *= scale;
            samples_extra[i] *= scale;
        }
    } else {
    /* Allow overshoot */
        for (i = start; i < end; i++) {
            phase = (double)i * 2.0 * M_PI * inv_sample_countf;
            samples_base[i] = 0.0f;
            samples_extra[i] = 0.0f;
            for (h = fh; h <= max_hb; h += hs) {
                hf = (double)h;
                samples_base[i] += (LADSPA_Data)((1.0 / hf) * sin(phase * hf));
            }
            for (; h <= max_he; h++) {
                hf = (double)h;
                samples_extra[i] = (LADSPA_Data)((1.0 / hf) * sin(phase * hf));
            }
            samples_base[i] *= scale;
            samples_extra[i] *= scale;
        }
    }
}

/*
 * Generate bandlimited square wave
 */
void
generateSquare(Wtblfile *wtblfile,
                 float gibbs_comp)
{
/* Set up to ignore first and last two samples */
    unsigned long sample_count = wtblfile->sample_count - 3;
    unsigned long start = 1, end = sample_count + 1;
    double inv_sample_countf = 1.0 / (double)sample_count;
    LADSPA_Data *samples_base = wtblfile->samples_base;
    LADSPA_Data *samples_extra = wtblfile->samples_extra;
    unsigned long max_hb = wtblfile->max_harmonic_base;
    unsigned long max_he = wtblfile->max_harmonic_extra;
    unsigned long fh = wave_first_harmonics[1];
    unsigned long hs = wave_harmonic_intervals[1];

    LADSPA_Data scale = 4.0f / M_PI;
    unsigned long i, h;
    double phase;
    double hf;
    double k;
    double m;

    if (gibbs_comp > 0.0f) {
    /* Compensate for Gibbs effect (overshoot at peaks) */
        k = M_PI * gibbs_comp / ((double)max_he * 2.0);
        for (i = start; i < end; i++) {
            samples_base[i] = 0.0f;
            samples_extra[i] = 0.0f;
        }
        for (h = fh; h <= max_hb; h += hs) {
            hf = (double)h;
            m = cos((hf - 1.0) * k);
            m *= m;
            for (i = start; i < end; i++) {
                phase = (double)i * 2.0 * M_PI * inv_sample_countf;
                samples_base[i] += (LADSPA_Data)((1.0 / hf) * sin(phase * hf) * m);
            }
        }
        for (; h <= max_he; h += hs) {
            hf = (double)h;
            m = cos((hf - 1.0) * k);
            m *= m;
            for (i = start; i < end; i++) {
                phase = (double)i * 2.0 * M_PI * inv_sample_countf;
                samples_extra[i] += (LADSPA_Data)((1.0 / hf) * sin(phase * hf) * m);
            }
        }
        for (i = start; i < end; i++) {
            samples_base[i] *= scale;
            samples_extra[i] *= scale;
        }
    } else {
        for (i = start; i < end; i++) {
            phase = (double)i * 2.0 * M_PI * inv_sample_countf;
            samples_base[i] = 0.0f;
            samples_extra[i] = 0.0f;
            for (h = fh; h <= max_hb; h += hs) {
                hf = (double)h;
                samples_base[i] += (LADSPA_Data)((1.0 / hf) * sin(phase * hf));
            }
            for (; h <= max_he; h++) {
                hf = (double)h;
                samples_extra[i] += (LADSPA_Data)((1.0 / hf) * sin(phase * hf));
            }
            samples_base[i] *= scale;
            samples_extra[i] *= scale;
        }
    }
}

/*
 * Generate bandlimited parabola wave
 */
void
generateParabola(Wtblfile *wtblfile,
                 float gibbs_comp)
{
/* Set up to ignore first and last two samples */
    unsigned long sample_count = wtblfile->sample_count - 3;
    unsigned long start = 1, end = sample_count + 1;
    double inv_sample_countf = 1.0 / (double)sample_count;
    LADSPA_Data *samples_base = wtblfile->samples_base;
    LADSPA_Data *samples_extra = wtblfile->samples_extra;
    unsigned long max_hb = wtblfile->max_harmonic_base;
    unsigned long max_he = wtblfile->max_harmonic_extra;
    unsigned long fh = wave_first_harmonics[2];
    unsigned long hs = wave_harmonic_intervals[2];

    LADSPA_Data scale = 2.0f / (M_PI * M_PI);
    unsigned long i, h;
    double phase;
    double hf;
    double k;
    double m;
    double sign;

    if (gibbs_comp > 0.0f) {
    /* Compensate for Gibbs effect (overshoot at peaks) */
        k = M_PI * gibbs_comp / ((double)max_he * 2.0);
        sign = -1.0;

        for (i = start; i < end; i++) {
            samples_base[i] = 0.0f;
            samples_extra[i] = 0.0f;
        }
        for (h = fh; h <= max_hb; h += hs) {
            hf = (double)h;
            m = cos((hf - 1.0) * k);
            m *= m;
            for (i = start; i < end; i++) {
                phase = (double)i * 2.0 * M_PI * inv_sample_countf;
                samples_base[i] += (LADSPA_Data)((sign * 4.0 / (hf * hf)) * cos(phase * hf) * m);
            }
            sign = -sign;
        }
        for (; h <= max_he; h += hs) {
            hf = (double)h;
            m = cos((hf - 1.0) * k);
            m *= m;
            for (i = start; i < end; i++) {
                phase = (double)i * 2.0 * M_PI * inv_sample_countf;
                samples_extra[i] += (LADSPA_Data)((sign * 4.0 / (hf * hf)) * cos(phase * hf) * m);
            }
            sign = -sign;
        }
        for (i = start; i < end; i++) {
            samples_base[i] *= scale;
            samples_extra[i] *= scale;
        }
    } else {
        for (i = start; i < end; i++) {
            phase = (double)i * 2.0 * M_PI * inv_sample_countf;
            samples_base[i] = 0.0f;
            samples_extra[i] = 0.0f;
            sign = -1.0;
            for (h = fh; h <= max_hb; h += hs) {
                hf = (double)h;
                samples_base[i] += (LADSPA_Data)((sign * 4.0 / (hf * hf)) * cos(phase * hf));
                sign = -sign;
            }
            for (; h <= max_he; h++) {
                hf = (double)h;
                samples_extra[i] += (LADSPA_Data)((sign * 4.0 / (hf * hf)) * cos(phase * hf));
                sign = -sign;
            }
            samples_base[i] *= scale;
            samples_extra[i] *= scale;
        }
    }
}

/*
 * Initialise a wavetable with a given waveform
 */
void
generateWave(Wtblfile *wtblfile,
             int wavetype_index,
             float gibbs_comp)
{
    LADSPA_Data *samples;
    unsigned long count;

    wave_functions[wavetype_index](wtblfile, gibbs_comp);

/*
 * The wave generation functions ignore the first and
 * last two samples. These are duplicates of samples in
 * the table, and are used for padding when
 * interpolating
 */
    count = wtblfile->sample_count - 3;

    samples = wtblfile->samples_base;
    samples[0] = samples[count];
    samples[count + 1] = samples[1];
    samples[count + 2] = samples[2];

    samples = wtblfile->samples_extra;
    samples[0] = samples[count];
    samples[count + 1] = samples[1];
    samples[count + 2] = samples[2];
}

/*
 * Allocate and intialise space for new wavedata file
 */
Wdatfile *
createWavedataFile(unsigned long sample_rate,
                   unsigned long lowest_harmonic,
                   unsigned long highest_harmonic)
{
    Wdatfile *wdatfile;

    wdatfile = (Wdatfile *)malloc(sizeof(Wdatfile));
    if (!wdatfile) {
        return NULL;
    }

    wdatfile->sample_rate = sample_rate;
    wdatfile->number_of_wavetables = 0;
    wdatfile->lowest_harmonic = lowest_harmonic;
    wdatfile->highest_harmonic = highest_harmonic;
    wdatfile->wavetables = NULL;

    return wdatfile;
}

/*
 * Free up all allocated data
 */
void
destroyWavedataFile(Wdatfile *wdatfile)
{
    unsigned long w;

    for (w = 0; w < wdatfile->number_of_wavetables; w++) {
        if (wdatfile->wavetables[w]) free(wdatfile->wavetables[w]);
    }
    if (wdatfile->wavetables) {
        free(wdatfile->wavetables);
    }
    free(wdatfile);
}

/*
 * Allocate and initialise new space for wavetable item in
 * Wavedata file object
 */
int
addWavetable(Wdatfile *wdatfile,
             unsigned long sample_count,
             unsigned long max_harmonic_base,
             unsigned long max_harmonic_extra)
{
    Wtblfile **wtblfiles;
    Wtblfile *wtblfile;
    LADSPA_Data *samples_base;
    LADSPA_Data *samples_extra;
    unsigned long w_count = wdatfile->number_of_wavetables;
    size_t bytesize;

    bytesize = (w_count + 1) * sizeof(Wtblfile *);
    if (!wdatfile->wavetables) {
        wtblfiles = (Wtblfile **)malloc(bytesize);
    } else {
        wtblfiles = (Wtblfile **)realloc(wdatfile->wavetables, bytesize);
    }

    if (!wtblfiles) {
        return FALSE;
    }

    wdatfile->wavetables = wtblfiles;
    wdatfile->number_of_wavetables++;

    wtblfile = (Wtblfile *)malloc(sizeof(Wtblfile));

    if (!wtblfile) {
        wtblfiles[w_count] = NULL;
        return FALSE;
    }

    wtblfiles[w_count] = wtblfile;

/* We add an extra 3 samples per table to allow for interpolation, which
   requires one previous sample, and two following */
    sample_count += 3;
    bytesize = (size_t)(sample_count) * sizeof(LADSPA_Data);
    samples_base = (LADSPA_Data *)malloc(bytesize);
    samples_extra = (LADSPA_Data *)malloc(bytesize);

    if (!samples_base || !samples_extra) {
        if (samples_base) free(samples_base);
        if (samples_extra) free(samples_extra);
        free(wtblfile);
        wtblfiles[w_count] = NULL;
        return FALSE;
    }

    wtblfile->sample_count = sample_count;
    wtblfile->max_harmonic_base = max_harmonic_base;
    wtblfile->max_harmonic_extra = max_harmonic_extra;
    wtblfile->samples_base = samples_base;
    wtblfile->samples_extra = samples_extra;

    return TRUE;
}

/*
 * Write wavedata file data to file
 */
int
writeWavedataFile(Wdatfile *wdatfile,
                  FILE *wdat_fp,
                  char *prefix)
{
    char *pfx;
    char *PFX;
    size_t pfxlen;
    unsigned long s_rate = wdatfile->sample_rate;
    unsigned long w, w_count = wdatfile->number_of_wavetables;
    unsigned long h_min = wdatfile->lowest_harmonic;
    unsigned long h_max = wdatfile->highest_harmonic;
    unsigned long s;
    int column;

    if (!wdat_fp) {
        return FALSE;
    }

    if (!prefix) {
        pfx = strdup("");
        PFX = strdup("");
    } else {
        pfxlen = strlen(prefix);
        pfx = malloc(pfxlen + 2);
        pfx = strncpy(pfx, prefix, pfxlen);
        pfx[pfxlen] = '_';
        pfx[pfxlen + 1] = '\0';
        PFX = strdup(pfx);
        pfxlen = strlen(pfx);
        while (pfxlen-- > 0) {
            PFX[pfxlen] = PFX[pfxlen] > 96 && PFX[pfxlen] < 123 ? PFX[pfxlen] - 32 : PFX[pfxlen];
        }
    }

    fprintf(wdat_fp, "#ifndef %sWAVEDATAFILE_H\n", PFX);
    fprintf(wdat_fp, "#define %sWAVEDATAFILE_H\n", PFX);
    fprintf(wdat_fp, "\n");
    fprintf(wdat_fp, "#include \"ladspa.h\"\n");
    fprintf(wdat_fp, "\n");
    fprintf(wdat_fp, "unsigned long %ssample_rate = %ld;\n", pfx, s_rate);
    fprintf(wdat_fp, "\n");
    fprintf(wdat_fp, "unsigned long %snumber_of_wavetables = %ld;\n", pfx, w_count);
    fprintf(wdat_fp, "\n");
    fprintf(wdat_fp, "unsigned long %slowest_harmonic = %ld;\n", pfx, h_min);
    fprintf(wdat_fp, "\n");
    fprintf(wdat_fp, "unsigned long %shighest_harmonic = %ld;\n", pfx, h_max);
    fprintf(wdat_fp, "\n");

    fprintf(wdat_fp, "unsigned long %smax_harmonics[] = {\n", pfx);
    column = 0;
    for (w = 0; w < w_count - 1; w++, column++) {
        if (column == 8) {
            fprintf(wdat_fp, "\n");
            column = 0;
        }
        fprintf(wdat_fp, "%ld,", wdatfile->wavetables[w]->max_harmonic_extra);
    }
    if (column == 8) {
        fprintf(wdat_fp, "\n");
        column = 0;
    }
    fprintf(wdat_fp, "%ld\n", wdatfile->wavetables[w]->max_harmonic_extra);
    fprintf(wdat_fp, "};\n");
    fprintf(wdat_fp, "\n");

    fprintf(wdat_fp, "unsigned long %ssample_counts[] = {\n", pfx);
    column = 0;
    for (w = 0; w < w_count - 1; w++, column++) {
        if (column == 8) {
            fprintf(wdat_fp, "\n");
            column = 0;
        }
        fprintf(wdat_fp, "%ld,", wdatfile->wavetables[w]->sample_count);
    }
    if (column == 8) {
        fprintf(wdat_fp, "\n");
    }
    fprintf(wdat_fp, "%ld\n", wdatfile->wavetables[w]->sample_count);
    fprintf(wdat_fp, "};\n");
    fprintf(wdat_fp, "\n");

    fprintf(wdat_fp, "LADSPA_Data %ssample_data_base[] = {\n", pfx);
    column = 0;
    for (w = 0; w < w_count - 1; w++) {
        for (s = 0; s < wdatfile->wavetables[w]->sample_count; s++, column++) {
            if (column == 5) {
                fprintf(wdat_fp, "\n");
                column = 0;
            }
            fprintf(wdat_fp, "%+.8ef,", wdatfile->wavetables[w]->samples_base[s]);
        }
    }
    for (s = 0; s < wdatfile->wavetables[w]->sample_count - 1; s++, column++) {
        if (column == 5) {
            fprintf(wdat_fp, "\n");
            column = 0;
        }
        fprintf(wdat_fp, "%+.8ef,", wdatfile->wavetables[w]->samples_base[s]);
    }
    if (column == 5) {
        fprintf(wdat_fp, "\n");
    }
    fprintf(wdat_fp, "%+.8ef\n", wdatfile->wavetables[w]->samples_base[s]);
    fprintf(wdat_fp, "};\n");
    fprintf(wdat_fp, "\n");

    fprintf(wdat_fp, "LADSPA_Data %ssample_data_extra[] = {\n", pfx);
    column = 0;
    for (w = 0; w < w_count - 1; w++) {
        for (s = 0; s < wdatfile->wavetables[w]->sample_count; s++, column++) {
            if (column == 5) {
                fprintf(wdat_fp, "\n");
                column = 0;
            }
            fprintf(wdat_fp, "%+.8ef,", wdatfile->wavetables[w]->samples_extra[s]);
        }
    }
    for (s = 0; s < wdatfile->wavetables[w]->sample_count - 1; s++, column++) {
        if (column == 5) {
            fprintf(wdat_fp, "\n");
            column = 0;
        }
        fprintf(wdat_fp, "%+.8ef,", wdatfile->wavetables[w]->samples_extra[s]);
    }
    if (column == 5) {
        fprintf(wdat_fp, "\n");
    }
    fprintf(wdat_fp, "%+.8ef\n", wdatfile->wavetables[w]->samples_extra[s]);
    fprintf(wdat_fp, "};\n");
    fprintf(wdat_fp, "\n");

    fprintf(wdat_fp, "#endif /* %sWAVEDATAFILE_H */\n", PFX);

    free(pfx);

    return TRUE;
}
